# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import json
import logging

from grimoirelab_toolkit.datetime import str_to_datetime

from ..raw.elastic import PRJ_JSON_FILTER_SEPARATOR
from .enrich import Enrich, metadata, anonymize_url
from ..elastic_mapping import Mapping as BaseMapping

logger = logging.getLogger(__name__)


REVISION_TYPE = 'revision'
HIDDEN_EDITOR = '--hidden--'


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
                "revision_comment_analyzed": {
                    "type": "text",
                    "index": true
                },
                "title_analyzed": {
                    "type": "text",
                    "index": true
                }
           }
        } """

        return {"items": mapping}


class MediaWikiEnrich(Enrich):

    mapping = Mapping

    def get_field_unique_id(self):
        return "revision_revid"

    def get_identities(self, item):
        """ Return the identities from an item """
        identities = []

        if 'data' not in item:
            return identities
        if 'revisions' not in item['data']:
            return identities

        revisions = item['data']['revisions']

        for revision in revisions:
            user = self.get_sh_identity(revision)
            yield user

    def get_field_author(self):
        return 'user'

    def get_sh_identity(self, item, identity_field=None):

        if identity_field is None:
            identity_field = self.get_field_author()

        revision = item

        identity = {}
        identity['username'] = None
        identity['email'] = None
        identity['name'] = None

        if isinstance(item, dict) and 'data' in item:
            if 'revisions' not in item['data']:
                return identity
            # Use as identity the first reviewer for a page
            revision = item['data']['revisions'][0]

        if identity_field in revision:
            identity['username'] = revision[identity_field]
            identity['name'] = revision[identity_field]

        return identity

    def get_review_sh(self, revision, item):
        """ Add sorting hat enrichment fields for the author of the revision """

        identity = self.get_sh_identity(revision)
        update = str_to_datetime(item[self.get_field_date()])
        erevision = self.get_item_sh_fields(identity, update)

        return erevision

    def get_rich_item_reviews(self, item):
        erevisions = []
        eitem = {}

        # All revisions include basic page info
        eitem = self.get_rich_item(item)

        # Revisions
        if "revisions" not in item["data"]:
            return erevisions

        for rev in item["data"]["revisions"]:
            erevision = {}
            self.copy_raw_fields(self.RAW_FIELDS_COPY, eitem, erevision)
            # Metadata related to the page according to the enrichment specification
            copy_fields_item = ["origin", "metadata__updated_on", "metadata__timestamp", "pageid", "title"]
            for f in copy_fields_item:
                if f in eitem:
                    erevision["page_" + f] = eitem[f]
                else:
                    erevision["page_" + f] = None
            # Copy fields from the review
            copy_fields = ["revid", "user", "parentid", "timestamp", "comment"]
            for f in copy_fields:
                if f in rev:
                    erevision["revision_" + f] = rev[f]
                else:
                    erevision["revision_" + f] = None

            if "comment" in rev:
                erevision["revision_comment"] = rev["comment"][:self.KEYWORD_MAX_LENGTH]
                erevision["revision_comment_analyzed"] = rev["comment"]

            if self.sortinghat:
                erevision.update(self.get_review_sh(rev, item))

            if self.prjs_map:
                erevision.update(self.get_item_project(erevision))

            # And now some calculated fields
            if self.prjs_map and "mediawiki" in self.prjs_map:
                for repo in self.prjs_map["mediawiki"].keys():
                    repo = repo.split(PRJ_JSON_FILTER_SEPARATOR)[0].strip()
                    if erevision["page_origin"] in repo:
                        urls = repo.split()
                        # If only one URL is given, then we consider same URL for API and web server
                        if len(urls) == 1:
                            erevision["url"] = urls[0] + "/" + erevision["page_title"]
                        elif len(urls) == 2:
                            erevision["url"] = urls[1] + "/" + erevision["page_title"]
                        else:
                            raise ValueError("Parameter value not supported in projects.json for mediawiki: " + repo)
            else:
                erevision["url"] = erevision["page_origin"] + "/" + erevision["page_title"]

            erevision["url"] = erevision["url"].replace(" ", "_")
            erevision["iscreated"] = 0
            erevision["isrevision"] = 0
            erevision["creation_date"] = rev['timestamp']

            if rev['parentid'] == 0:
                erevision["iscreated"] = 1
            else:
                erevision["isrevision"] = 1

            erevision["page_last_edited_date"] = eitem['last_edited_date']

            erevision['metadata__gelk_version'] = eitem['metadata__gelk_version']
            erevision['metadata__gelk_backend_name'] = eitem['metadata__gelk_backend_name']
            erevision['metadata__enriched_on'] = eitem['metadata__enriched_on']
            erevision['repository_labels'] = eitem['repository_labels']

            erevision.update(self.get_grimoire_fields(erevision['creation_date'], REVISION_TYPE))

            yield erevision

    @metadata
    def get_rich_item(self, item):
        eitem = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)
        # The real data
        page = item['data']

        # data fields to copy
        copy_fields = ["pageid", "title"]
        for f in copy_fields:
            if f in page:
                eitem[f] = page[f]
            else:
                eitem[f] = None
        # Fields which names are translated
        map_fields = {"title": "title_analyzed"}
        for fn in map_fields:
            eitem[map_fields[fn]] = page[fn]

        # Enrich dates
        eitem["update_date"] = str_to_datetime(item["metadata__updated_on"]).isoformat()
        # Revisions
        eitem["last_edited_date"] = None
        eitem["nrevisions"] = 0
        if "revisions" in page:
            eitem["nrevisions"] = len(page["revisions"])
            if len(page["revisions"]) > 0:
                eitem["first_editor"] = page["revisions"][0].get('user', HIDDEN_EDITOR)
                eitem["last_edited_date"] = page["revisions"][-1]["timestamp"]

        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)
        return eitem

    def enrich_items(self, items):
        # Enrich always events for MediaWiki items
        return self.enrich_events(items)

    def enrich_events(self, ocean_backend):
        max_items = self.elastic.max_items_bulk
        current = 0
        bulk_json = ""
        total = 0

        url = self.elastic.get_bulk_url()

        logger.debug("[mediawiki] Adding items to {} (in {} packs)".format(anonymize_url(url), max_items))

        items = ocean_backend.fetch()
        for item in items:
            rich_item_reviews = self.get_rich_item_reviews(item)
            for enrich_review in rich_item_reviews:
                if current >= max_items:
                    total += self.elastic.safe_put_bulk(url, bulk_json)
                    bulk_json = ""
                    current = 0
                data_json = json.dumps(enrich_review)
                bulk_json += '{"index" : {"_id" : "%s" } }\n' % \
                    (enrich_review[self.get_field_unique_id()])
                bulk_json += data_json + "\n"  # Bulk document
                current += 1

        if current > 0:
            total += self.elastic.safe_put_bulk(url, bulk_json)

        return total



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jose Javier Merchante Picazo <jjmerchante@cauldron.io>
#

from grimoire_elk.identities.identities import Identities


class StackExchangeIdentities(Identities):
    @classmethod
    def anonymize_item(cls, item):
        """Remove or hash the fields that contain personal information

        Comments are removed because can cause complexity, there could be many
        and are not used in the enrichment process
        """

        item = item['data']

        item['comments'] = []
        if 'owner' in item and item['owner']:
            cls._sanitize_owner(item['owner'])

        if 'answers' in item and item['answers']:
            for answer in item['answers']:
                if 'owner' in answer and answer['owner']:
                    cls._sanitize_owner(answer['owner'])
                answer['comments'] = []

    @classmethod
    def _sanitize_owner(cls, owner):
        """Remove links and hash personal information"""
        if 'display_name' in owner:
            owner['display_name'] = cls._hash(owner['display_name'])
        if 'user_id' in owner:
            owner['user_id'] = cls._hash(str(owner['user_id']))
        owner['profile_image'] = ''
        owner['link'] = ''



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Alberto Perez Garcia-Plaza <alpgarcia@bitergia.com>
#

import configparser
import json

import httpretty
import requests
import sys
import unittest
from unittest.mock import MagicMock

from grimoire_elk.elastic import logger
from grimoire_elk.enriched.enrich import (Enrich,
                                          HEADER_JSON,
                                          anonymize_url)
from grimoire_elk.utils import get_connectors, get_elastic

# Make sure we use our code and not any other could we have installed
sys.path.insert(0, '..')

CONFIG_FILE = 'tests.conf'


def setup_http_server(url, not_handle_status_code=False):
    """Setup a mock HTTP server"""

    http_requests = []

    body_content_1 = read_file("data/author_min_max_dates_1.json")
    body_content_2 = read_file("data/author_min_max_dates_2.json")
    body_content_empty = read_file("data/author_min_max_dates_empty.json")

    def request_callback(method, uri, headers):

        status_code = 200
        composite = method.parsed_body['aggs']['author']['composite']
        if "after" in composite:
            if composite["after"]["author_uuid"] == "007a56d0322c518859dde2a0c6ed9143fa141c61":
                body = body_content_2
            else:
                body = body_content_empty
        else:
            body = body_content_1
        http_requests.append(httpretty.last_request())

        return status_code, headers, body

    httpretty.register_uri(httpretty.POST,
                           url,
                           match_querystring=True,
                           responses=[
                               httpretty.Response(body=request_callback)
                           ])

    return http_requests


def read_file(filename):
    with open(filename) as f:
        return f.read()


class TestEnrich(unittest.TestCase):

    def setUp(self):
        config = configparser.ConfigParser()
        config.read(CONFIG_FILE)
        es_con = dict(config.items('ElasticSearch'))['url']
        git_enriched = "git_enriched"
        enrich_backend = get_connectors()["git"][2]()
        elastic_enrich = get_elastic(es_con, git_enriched, True, enrich_backend)
        self._enrich = Enrich()
        self._enrich.elastic = elastic_enrich

        self.empty_item = {
            "author_id": "",
            "author_uuid": "",
            "author_name": "",
            "author_user_name": "",
            "author_domain": "",
            "author_gender": "",
            "author_gender_acc": None,
            "author_org_name": "",
            "author_bot": False,
            "author_multi_org_names": [""]
        }

    def test_get_item_sh_fields_identity(self):
        """Test sortinghat fields retrieval for a given identity"""

        # Method Parameter: Identity
        identity = {
            "name": "pepe",
            "username": "pepepotamo"
        }

        expected = {
            'author_id': "11111",
            'author_uuid': "aaaaa",
            'author_name': "Pepe",
            'author_user_name': identity["username"],
            'author_domain': "local",
            'author_gender': "male",
            'author_gender_acc': 100,
            'author_org_name': "Bitergia",
            'author_bot': False,
            'author_multi_org_names': ['Bitergia']
        }

        empty_item_by_rol = {'author': self.empty_item}

        def rol_side_effect(rol):
            return empty_item_by_rol[rol]
        self._enrich.__get_item_sh_fields_empty = MagicMock(side_effect=rol_side_effect)

        entity = {
            'id': expected['author_id'],
            'uuid': expected['author_uuid'],
            'name': expected['author_name'],
            'username': identity["username"],
            'email': None,  # None to verify email is from profile
            'profile': {
                "name": expected['author_name'],
                "email": "pepepotamo@local.com",
                "gender": expected['author_gender'],
                "genderAcc": expected['author_gender_acc'],
                "isBot": expected['author_bot']
            },
            'enrollments': [
                {
                    'group': {
                        'parent_org': None,
                        'name': expected['author_org_name'],
                        'type': 'organization',
                        'start': "1900-01-01T00:00:00+00:00",
                        'end': "2100-01-01T00:00:00+00:00"
                    }
                }
            ]
        }
        self._enrich.get_sh_item_from_identity = MagicMock(return_value=entity)

        email_doms = {'pepepotamo@local.com': expected['author_domain']}

        def email_domain_side_effect(email):
            return email_doms[email]
        self._enrich.get_email_domain = MagicMock(side_effect=email_domain_side_effect)

        # 1. Author role

        # Method to test
        eitem_sh = self._enrich.get_item_sh_fields(identity=identity)
        self.assertEqual(eitem_sh['author_id'], expected['author_id'])
        self.assertEqual(eitem_sh['author_uuid'], expected['author_uuid'])
        self.assertEqual(eitem_sh['author_name'], expected['author_name'])
        self.assertEqual(eitem_sh['author_user_name'], expected['author_user_name'])
        self.assertEqual(eitem_sh['author_domain'], expected['author_domain'])
        self.assertEqual(eitem_sh['author_gender'], expected['author_gender'])
        self.assertEqual(eitem_sh['author_gender_acc'], expected['author_gender_acc'])
        self.assertEqual(eitem_sh['author_org_name'], expected['author_org_name'])
        self.assertEqual(eitem_sh['author_bot'], expected['author_bot'])
        self.assertEqual(eitem_sh['author_multi_org_names'], expected['author_multi_org_names'])

        # 2. Change role

        # Method to test
        eitem_sh = self._enrich.get_item_sh_fields(identity=identity, rol='assignee')
        self.assertEqual(eitem_sh['assignee_id'], expected['author_id'])
        self.assertEqual(eitem_sh['assignee_uuid'], expected['author_uuid'])
        self.assertEqual(eitem_sh['assignee_name'], expected['author_name'])
        self.assertEqual(eitem_sh['assignee_user_name'], expected['author_user_name'])
        self.assertEqual(eitem_sh['assignee_domain'], expected['author_domain'])
        self.assertEqual(eitem_sh['assignee_gender'], expected['author_gender'])
        self.assertEqual(eitem_sh['assignee_gender_acc'], expected['author_gender_acc'])
        self.assertEqual(eitem_sh['assignee_org_name'], expected['author_org_name'])
        self.assertEqual(eitem_sh['assignee_bot'], expected['author_bot'])
        self.assertEqual(eitem_sh['assignee_multi_org_names'], expected['author_multi_org_names'])

    def test_get_item_sh_fields_identity_no_uuid(self):
        """uuid is None (not found in sortinghat) or does not exist"""

        # Method Parameter: Identity
        identity = {
            "name": "pepe",
            "username": "pepepotamo"
        }

        # In this test 'empty_item' is also the expected result

        # 1. uuid is an empty string

        empty_item_by_rol = {'author': self.empty_item}

        def rol_side_effect(rol):
            return empty_item_by_rol[rol]
        self._enrich.__get_item_sh_fields_empty = MagicMock(side_effect=rol_side_effect)

        sh_ids = {
            "id": "",
            "uuid": ""
        }
        self._enrich.get_sh_item_from_identity = MagicMock(return_value=sh_ids)

        # Method to test
        eitem_sh = self._enrich.get_item_sh_fields(identity=identity)
        self.assertEqual(eitem_sh['author_id'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_uuid'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_name'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_user_name'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_domain'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_gender'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_gender_acc'], None)
        self.assertEqual(eitem_sh['author_org_name'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_multi_org_names'], ['-- UNDEFINED --'])

        # 2. No uuid field
        sh_ids = {
            "id": ""
        }
        self._enrich.get_sh_item_from_identity = MagicMock(return_value=sh_ids)

        # Method to test
        eitem_sh = self._enrich.get_item_sh_fields(identity=identity)
        self.assertEqual(eitem_sh['author_id'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_uuid'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_name'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_user_name'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_domain'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_gender'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_gender_acc'], None)
        self.assertEqual(eitem_sh['author_org_name'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_bot'], False)
        self.assertEqual(eitem_sh['author_multi_org_names'], ['-- UNDEFINED --'])

    def test_get_item_sh_fields_identity_no_profile(self):
        """Test retrieval when no profile data is found or data is not what we expected"""

        # Method Parameter: Identity
        identity = {
            "name": "pepe",
            "username": "pepepotamo"
        }

        # As there is no profile, name should come from the identity,
        # domain should be None as specified above in self._enrich.get_identity_domain
        # mock, and gender fields will have values coming from the
        # initial empty item.
        expected = {
            'author_id': "11111",
            'author_uuid': "aaaaa",
            'author_name': "pepe",
            'author_user_name': identity["username"],
            'author_domain': None,
            'author_gender': self._enrich.unknown_gender,
            'author_gender_acc': 0,
            'author_org_name': "Bitergia",
            'author_bot': False,
            'author_multi_org_names': ['Bitergia']
        }

        # 1. Profile is None

        empty_item_by_rol = {'author': self.empty_item}

        def rol_side_effect(rol):
            return empty_item_by_rol[rol]
        self._enrich.__get_item_sh_fields_empty = MagicMock(side_effect=rol_side_effect)

        entity = {
            'id': expected['author_id'],
            'uuid': expected['author_uuid'],
            'name': expected['author_name'],
            'username': identity["username"],
            'email': None,  # None to verify email is from profile
            'profile': None,
            'enrollments': [
                {
                    'group': {
                        'parent_org': None,
                        'name': expected['author_org_name'],
                        'type': 'organization',
                        'start': "1900-01-01T00:00:00+00:00",
                        'end': "2100-01-01T00:00:00+00:00"
                    }
                }
            ]
        }
        self._enrich.get_sh_item_from_identity = MagicMock(return_value=entity)

        email_doms = {'pepepotamo@local.com': expected['author_domain']}

        def email_domain_side_effect(email):
            return email_doms[email]
        self._enrich.get_email_domain = MagicMock(side_effect=email_domain_side_effect)

        # Method to test
        eitem_sh = self._enrich.get_item_sh_fields(identity=identity)
        self.assertEqual(eitem_sh['author_id'], expected['author_id'])
        self.assertEqual(eitem_sh['author_uuid'], expected['author_uuid'])
        self.assertEqual(eitem_sh['author_name'], expected['author_name'])
        self.assertEqual(eitem_sh['author_user_name'], expected['author_user_name'])
        self.assertEqual(eitem_sh['author_domain'], expected['author_domain'])
        self.assertEqual(eitem_sh['author_gender'], expected['author_gender'])
        self.assertEqual(eitem_sh['author_gender_acc'], expected['author_gender_acc'])
        self.assertEqual(eitem_sh['author_org_name'], expected['author_org_name'])
        self.assertEqual(eitem_sh['author_bot'], expected['author_bot'])
        self.assertEqual(eitem_sh['author_multi_org_names'], expected['author_multi_org_names'])

        # 2. Profile as empty dict

        entity['profile'] = {}

        # Method to test
        eitem_sh = self._enrich.get_item_sh_fields(identity=identity)
        # Same as 1.3.
        self.assertEqual(eitem_sh['author_id'], expected['author_id'])
        self.assertEqual(eitem_sh['author_uuid'], expected['author_uuid'])
        self.assertEqual(eitem_sh['author_name'], expected['author_name'])
        self.assertEqual(eitem_sh['author_user_name'], expected['author_user_name'])
        self.assertEqual(eitem_sh['author_domain'], expected['author_domain'])
        self.assertEqual(eitem_sh['author_gender'], expected['author_gender'])
        self.assertEqual(eitem_sh['author_gender_acc'], expected['author_gender_acc'])
        self.assertEqual(eitem_sh['author_org_name'], expected['author_org_name'])
        self.assertEqual(eitem_sh['author_bot'], expected['author_bot'])
        self.assertEqual(eitem_sh['author_multi_org_names'], expected['author_multi_org_names'])

        # 3 Profile with other fields

        entity['profile'] = {
            "first": expected['author_name'],
            "second": "pepepotamo@local.com",
        }

        # Method to test
        self.assertEqual(eitem_sh['author_id'], expected['author_id'])
        self.assertEqual(eitem_sh['author_uuid'], expected['author_uuid'])
        self.assertEqual(eitem_sh['author_name'], expected['author_name'])
        self.assertEqual(eitem_sh['author_user_name'], expected['author_user_name'])
        self.assertEqual(eitem_sh['author_domain'], expected['author_domain'])
        self.assertEqual(eitem_sh['author_gender'], expected['author_gender'])
        self.assertEqual(eitem_sh['author_gender_acc'], expected['author_gender_acc'])
        self.assertEqual(eitem_sh['author_org_name'], expected['author_org_name'])
        self.assertEqual(eitem_sh['author_bot'], expected['author_bot'])
        self.assertEqual(eitem_sh['author_multi_org_names'], expected['author_multi_org_names'])

    def test_get_item_sh_fields_identity_no_gender(self):
        """Profile with no gender values"""

        # Method Parameter: Identity
        identity = {
            "name": "pepe",
            "username": "pepepotamo"
        }

        expected = {
            'author_id': "11111",
            'author_uuid': "aaaaa",
            'author_name': "Pepe",
            'author_user_name': identity["username"],
            'author_domain': "local",
            'author_gender': self._enrich.unknown_gender,
            'author_gender_acc': 0,
            'author_org_name': "Bitergia",
            'author_bot': False,
            'author_multi_org_names': ['Bitergia']
        }

        empty_item_by_rol = {'author': self.empty_item}

        def rol_side_effect(rol):
            return empty_item_by_rol[rol]
        self._enrich.__get_item_sh_fields_empty = MagicMock(side_effect=rol_side_effect)

        entity = {
            'id': expected['author_id'],
            'uuid': expected['author_uuid'],
            'name': expected['author_name'],
            'username': identity["username"],
            'email': None,  # None to verify email is from profile
            'profile': {
                "name": expected['author_name'],
                "email": "pepepotamo@local.com",
                "isBot": expected['author_bot']
            },
            'enrollments': [
                {
                    'group': {
                        'parent_org': None,
                        'name': expected['author_org_name'],
                        'type': 'organization',
                        'start': "1900-01-01T00:00:00+00:00",
                        'end': "2100-01-01T00:00:00+00:00"
                    }
                }
            ]
        }
        self._enrich.get_sh_item_from_identity = MagicMock(return_value=entity)

        email_doms = {'pepepotamo@local.com': expected['author_domain']}

        def email_domain_side_effect(email):
            return email_doms[email]
        self._enrich.get_email_domain = MagicMock(side_effect=email_domain_side_effect)

        # Method to test
        eitem_sh = self._enrich.get_item_sh_fields(identity=identity)
        self.assertEqual(eitem_sh['author_id'], expected['author_id'])
        self.assertEqual(eitem_sh['author_uuid'], expected['author_uuid'])
        self.assertEqual(eitem_sh['author_name'], expected['author_name'])
        self.assertEqual(eitem_sh['author_user_name'], expected['author_user_name'])
        self.assertEqual(eitem_sh['author_domain'], expected['author_domain'])
        self.assertEqual(eitem_sh['author_gender'], expected['author_gender'])
        self.assertEqual(eitem_sh['author_gender_acc'], expected['author_gender_acc'])
        self.assertEqual(eitem_sh['author_org_name'], expected['author_org_name'])
        self.assertEqual(eitem_sh['author_bot'], expected['author_bot'])
        self.assertEqual(eitem_sh['author_multi_org_names'], expected['author_multi_org_names'])

    def test_get_item_sh_fields_sh_id(self):
        """Test retrieval from sortinghat id"""

        sh_id = "11111"

        expected = {
            'author_id': sh_id,
            'author_uuid': "aaaaa",
            'author_name': "Pepe",
            'author_user_name': "",
            'author_domain': "local",
            'author_gender': "male",
            'author_gender_acc': 100,
            'author_org_name': "Bitergia",
            'author_bot': False,
            'author_multi_org_names': ['Bitergia']
        }

        empty_item_by_rol = {'author': self.empty_item}

        def rol_side_effect(rol):
            return empty_item_by_rol[rol]
        self._enrich.__get_item_sh_fields_empty = MagicMock(side_effect=rol_side_effect)

        entity = {
            'id': expected['author_id'],
            'uuid': expected['author_uuid'],
            'profile': {
                "name": expected['author_name'],
                "email": "pepepotamo@local.com",
                "gender": expected['author_gender'],
                "genderAcc": expected['author_gender_acc'],
                "isBot": expected['author_bot']
            },
            'enrollments': [
                {
                    'group': {
                        'parent_org': None,
                        'name': expected['author_org_name'],
                        'type': 'organization',
                        'start': "1900-01-01T00:00:00+00:00",
                        'end': "2100-01-01T00:00:00+00:00"
                    }
                }
            ]
        }
        self._enrich.get_sh_item_from_id = MagicMock(return_value=entity)

        email_doms = {'pepepotamo@local.com': expected['author_domain']}

        def email_domain_side_effect(email):
            return email_doms[email]
        self._enrich.get_email_domain = MagicMock(side_effect=email_domain_side_effect)

        # Method to test
        eitem_sh = self._enrich.get_item_sh_fields(sh_id=sh_id)
        self.assertEqual(eitem_sh['author_id'], expected['author_id'])
        self.assertEqual(eitem_sh['author_uuid'], expected['author_uuid'])
        self.assertEqual(eitem_sh['author_name'], expected['author_name'])
        self.assertEqual(eitem_sh['author_user_name'], expected['author_user_name'])
        self.assertEqual(eitem_sh['author_domain'], expected['author_domain'])
        self.assertEqual(eitem_sh['author_gender'], expected['author_gender'])
        self.assertEqual(eitem_sh['author_gender_acc'], expected['author_gender_acc'])
        self.assertEqual(eitem_sh['author_org_name'], expected['author_org_name'])
        self.assertEqual(eitem_sh['author_bot'], expected['author_bot'])
        self.assertEqual(eitem_sh['author_org_name'], expected['author_org_name'])
        self.assertEqual(eitem_sh['author_bot'], expected['author_bot'])
        self.assertEqual(eitem_sh['author_multi_org_names'], expected['author_multi_org_names'])

    def test_get_item_sh_fields_sh_id_no_uuid(self):
        """Test retrieval from sortinghat id when there is no uuid"""

        sh_id = "11111"

        # 1. uuid is None

        empty_item_by_rol = {'author': self.empty_item}

        def rol_side_effect(rol):
            return empty_item_by_rol[rol]
        self._enrich.__get_item_sh_fields_empty = MagicMock(side_effect=rol_side_effect)

        entity = {
            'id': sh_id,
            'uuid': None
        }
        self._enrich.get_sh_item_from_id = MagicMock(return_value=entity)

        # Method to test
        eitem_sh = self._enrich.get_item_sh_fields(sh_id=sh_id)
        self.assertEqual(eitem_sh['author_id'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_uuid'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_name'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_user_name'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_domain'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_gender'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_gender_acc'], None)
        self.assertEqual(eitem_sh['author_org_name'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_bot'], False)
        self.assertEqual(eitem_sh['author_multi_org_names'], ['-- UNDEFINED --'])

        # 2. uuid is an empty string

        entity['uuid'] = ""

        # Method to test
        eitem_sh = self._enrich.get_item_sh_fields(sh_id=sh_id)
        self.assertEqual(eitem_sh['author_id'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_uuid'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_name'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_user_name'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_domain'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_gender'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_gender_acc'], None)
        self.assertEqual(eitem_sh['author_org_name'], '-- UNDEFINED --')
        self.assertEqual(eitem_sh['author_bot'], False)
        self.assertEqual(eitem_sh['author_multi_org_names'], ['-- UNDEFINED --'])

    def test_get_main_enrollments(self):
        """Test get the main enrollment given the list of enrollments"""
        enrollments = ['Bitergia::Eng', 'Chaoss']
        main_enrolls = self._enrich.get_main_enrollments(enrollments)
        self.assertListEqual(main_enrolls, ['Bitergia', 'Chaoss'])

    def test_remove_prefix_enrollments(self):
        """Test remove the prefix enrollment given the list of enrollments"""
        enrollments = ['Bitergia::Eng', 'Chaoss']
        enrolls = self._enrich.remove_prefix_enrollments(enrollments)
        self.assertListEqual(enrolls, ['Chaoss', 'Eng'])

    def test_no_params(self):
        """Neither identity nor sh_id are passed as arguments"""

        empty_item_by_rol = {'author': self.empty_item}

        def rol_side_effect(rol):
            return empty_item_by_rol[rol]
        self._enrich.__get_item_sh_fields_empty = MagicMock(side_effect=rol_side_effect)

        # Method to test
        eitem_sh = self._enrich.get_item_sh_fields()
        self.assertEqual(eitem_sh['author_id'], self.empty_item['author_id'])
        self.assertEqual(eitem_sh['author_uuid'], self.empty_item['author_uuid'])
        self.assertEqual(eitem_sh['author_name'], self.empty_item['author_name'])
        self.assertEqual(eitem_sh['author_user_name'], self.empty_item['author_user_name'])
        self.assertEqual(eitem_sh['author_domain'], self.empty_item['author_domain'])
        self.assertEqual(eitem_sh['author_gender'], self.empty_item['author_gender'])
        self.assertEqual(eitem_sh['author_gender_acc'], self.empty_item['author_gender_acc'])
        self.assertEqual(eitem_sh['author_org_name'], self.empty_item['author_org_name'])
        self.assertEqual(eitem_sh['author_bot'], self.empty_item['author_bot'])
        self.assertEqual(eitem_sh['author_multi_org_names'], self.empty_item['author_multi_org_names'])

    def test_add_alias(self):
        """Test whether add_alias properly works"""

        config = configparser.ConfigParser()
        config.read(CONFIG_FILE)
        es_con = dict(config.items('ElasticSearch'))['url']

        tmp_index = "test-add-aliases"
        tmp_index_url = es_con + "/" + tmp_index

        enrich_backend = get_connectors()["git"][2]()
        elastic_enrich = get_elastic(es_con, tmp_index, True, enrich_backend)
        self._enrich.set_elastic(elastic_enrich)

        # add alias
        alias = "demographics"
        with self.assertLogs(logger, level='INFO') as cm:
            self._enrich.elastic.add_alias(alias)

        self.assertEqual(cm.output[0],
                         'INFO:grimoire_elk.elastic:Alias %s created on %s.'
                         % (alias, anonymize_url(tmp_index_url)))

        r = self._enrich.requests.get(self._enrich.elastic.index_url + "/_alias", headers=HEADER_JSON, verify=False)
        self.assertIn(alias, r.json()[self._enrich.elastic.index]['aliases'])

        # add alias again
        with self.assertLogs(logger, level='DEBUG') as cm:
            self._enrich.elastic.add_alias(alias)

        self.assertEqual(cm.output[0],
                         'DEBUG:grimoire_elk.elastic:Alias %s already exists on %s.'
                         % (alias, anonymize_url(tmp_index_url)))

        requests.delete(tmp_index_url, verify=False)

    def test_copy_raw_fields(self):
        """Test whether raw fields are copied properly"""

        raw_fields = ["metadata__updated_on", "metadata__timestamp",
                      "offset", "origin", "tag", "uuid", "extra"]
        source = {
            "metadata__updated_on": "2020-03-19T18:08:45.480000+00:00",
            "metadata__timestamp": "2020-04-11T12:45:37.229535+00:00",
            "offset": None,
            "origin": "https://gitter.im/jenkinsci/jenkins",
            "tag": "https://gitter.im/jenkinsci/jenkins",
            "uuid": "a9f92bf99785ad838df6736b7825c6f7ae16ac58",
        }

        expected = {
            "metadata__updated_on": "2020-03-19T18:08:45.480000+00:00",
            "metadata__timestamp": "2020-04-11T12:45:37.229535+00:00",
            "offset": None,
            "origin": "https://gitter.im/jenkinsci/jenkins",
            "tag": "https://gitter.im/jenkinsci/jenkins",
            "uuid": "a9f92bf99785ad838df6736b7825c6f7ae16ac58",
            "extra": None
        }

        eitem = {}
        self._enrich.copy_raw_fields(raw_fields, source, eitem)

        self.assertEqual(eitem['metadata__updated_on'], expected['metadata__updated_on'])
        self.assertEqual(eitem['metadata__timestamp'], expected['metadata__timestamp'])
        self.assertEqual(eitem['offset'], expected['offset'])
        self.assertEqual(eitem['origin'], expected['origin'])
        self.assertEqual(eitem['tag'], expected['tag'])
        self.assertEqual(eitem['uuid'], expected['uuid'])
        self.assertEqual(eitem['extra'], expected['extra'])

    def test_authors_min_max_dates(self):
        expected_es_query = """
        {
          "size": 0,
          "aggs": {
            "author": {
              "composite": {
                "sources": [
                  {
                    "author_uuid": {
                      "terms": {
                        "field": "author_uuid"
                      }
                    }
                  }
                ],
                "size": 10000
              },
              "aggs": {
                "min": {
                  "min": {
                    "field": "grimoire_creation_date"
                  }
                },
                "max": {
                  "max": {
                    "field": "grimoire_creation_date"
                  }
                }
              }
            }
          }
        }
        """
        es_query = self._enrich.authors_min_max_dates("grimoire_creation_date",
                                                      author_field="author_uuid",
                                                      contribution_type=None,
                                                      after=None)
        self.assertDictEqual(json.loads(es_query), json.loads(expected_es_query))

        expected_es_query = """
        {
          "size": 0,
          "aggs": {
            "author": {
              "composite": {
                "sources": [
                  {
                    "author_uuid": {
                      "terms": {
                        "field": "author_uuid"
                      }
                    }
                  }
                ],
                "after": {
                  "author_uuid": "uuid"
                },
                "size": 10000
              },
              "aggs": {
                "min": {
                  "min": {
                    "field": "grimoire_creation_date"
                  }
                },
                "max": {
                  "max": {
                    "field": "grimoire_creation_date"
                  }
                }
              }
            }
          }
        }
        """
        es_query = self._enrich.authors_min_max_dates("grimoire_creation_date",
                                                      author_field="author_uuid",
                                                      contribution_type=None,
                                                      after="uuid")
        self.assertDictEqual(json.loads(es_query), json.loads(expected_es_query))

    @httpretty.activate
    def test_fetch_authors_min_max_dates(self):

        es_search_url = "{}/_search".format(self._enrich.elastic.index_url)
        _ = setup_http_server(es_search_url)

        log_prefix = "[git] Demography"
        author_field = "author_uuid"
        date_field = "grimoire_creation_date"

        expected = [
            {
                'key': {'author_uuid': '00032fabbbf033467d7bd307df81b654c0fa53d8'},
                'doc_count': 1,
                'min': {'value': 1623225379000.0, 'value_as_string': '2021-06-09T07:56:19.000Z'},
                'max': {'value': 1623225379000.0, 'value_as_string': '2021-06-09T07:56:19.000Z'}
            },
            {
                'key': {'author_uuid': '007a56d0322c518859dde2a0c6ed9143fa141c61'},
                'doc_count': 1,
                'min': {'value': 1626183289000.0, 'value_as_string': '2021-07-13T13:34:49.000Z'},
                'max': {'value': 1626183289000.0, 'value_as_string': '2021-07-13T13:34:49.000Z'}
            },
            {
                'key': {'author_uuid': '00cc95a5950523a42c969f15c7c36c4530417f13'},
                'doc_count': 1,
                'min': {'value': 1474160034000.0, 'value_as_string': '2016-09-18T00:53:54.000Z'},
                'max': {'value': 1474160034000.0, 'value_as_string': '2016-09-18T00:53:54.000Z'}
            },
            {
                'key': {'author_uuid': '00d36515f739794b941586e5d0a102b5ff3a0cc2'},
                'doc_count': 1,
                'min': {'value': 1526521972000.0, 'value_as_string': '2018-05-17T01:52:52.000Z'},
                'max': {'value': 1526521972000.0, 'value_as_string': '2018-05-17T01:52:52.000Z'}
            }
        ]
        authors_min_max_data = self._enrich.fetch_authors_min_max_dates(log_prefix, author_field,
                                                                        None, date_field)
        all_authors = []
        for author_key in authors_min_max_data:
            all_authors.append(author_key)
        self.assertListEqual(all_authors, expected)

    def test_get_field_unique_id(self):
        self.assertEqual(self._enrich.get_field_unique_id(), 'uuid')

    def test_get_field_event_unique_id(self):
        with self.assertRaises(NotImplementedError):
            self._enrich.get_field_event_unique_id()

    def test_get_rich_item(self):
        with self.assertRaises(NotImplementedError):
            self._enrich.get_rich_item({})

    def test_get_rich_events(self):
        with self.assertRaises(NotImplementedError):
            self._enrich.get_rich_events({})

    def test_get_field_author(self):
        with self.assertRaises(NotImplementedError):
            self._enrich.get_field_author()

    def test_get_field_date(self):
        self.assertEqual(self._enrich.get_field_date(), 'metadata__updated_on')

    def test_get_identities(self):
        with self.assertRaises(NotImplementedError):
            self._enrich.get_identities(item=None)

    def test_add_repository_labels(self):
        item = {}
        self._enrich.add_repository_labels(item)
        self.assertIsNone(item['repository_labels'])

    def test_add_metadata_filter_raw(self):
        item = {}
        self._enrich.add_metadata_filter_raw(item)
        self.assertIsNone(item['metadata__filter_raw'])

    def test_get_item_id(self):
        item = {
            "_id": "0000"
        }
        self.assertEqual(self._enrich.get_item_id(item), item['_id'])


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#   Quan Zhou <quan@bitergia.com>
#

from datetime import datetime
import logging
import time

from sortinghat.cli.client import (SortingHatClientError,
                                   SortingHatSchema)
from sgqlc.operation import Operation


logger = logging.getLogger(__name__)

SLEEP_TIME = 5

DEFAULT_FROM_DATE = datetime(1900, 1, 1)
DEFAULT_TO_DATE = datetime(2100, 1, 1)

MULTI_ORG_NAMES = '_multi_org_names'

PAGE = 1
PAGE_SIZE = 100


class SortingHat(object):

    # Mutations
    @classmethod
    def add_enrollment(cls, db, uuid, organization, from_date=DEFAULT_FROM_DATE, to_date=DEFAULT_TO_DATE):
        args = {
            "from_date": from_date,
            "group": organization,
            "to_date": to_date,
            "uuid": uuid
        }
        try:
            op = Operation(SortingHatSchema.SortingHatMutation)
            add_org = op.enroll(**args)
            add_org.individual().enrollments().group().name()
            result = db.execute(op)
            org_name = result['data']['enroll']['individual']['enrollments'][0]['group']['name']
        except SortingHatClientError as e:
            logger.error("[sortinghat] Error add enrollment {} {} {} {}: {}".
                         format(uuid, organization, from_date, to_date, e.errors[0]['message']))
            org_name = None

        return org_name

    @classmethod
    def add_id(cls, db, source, email=None, name=None, username=None):
        uuid = None
        args = {
            "email": email,
            "name": name,
            "source": source,
            "username": username
        }
        args_without_empty = {k: v for k, v in args.items() if v}
        try:
            op = Operation(SortingHatSchema.SortingHatMutation)
            add = op.add_identity(**args_without_empty)
            add.uuid()
            result = db.execute(op)
            uuid = result['data']['addIdentity']['uuid']

            logger.debug("[sortinghat] New identity {} {},{},{} ".format(
                uuid, args['username'], args['name'], args['email']))
        except SortingHatClientError as ex:
            msg = ex.errors[0]['message']
            if 'already exists' in msg:
                logger.debug("[sortinghat] {}".format(msg))
                uuid = msg.split("'")[1]
            else:
                raise SortingHatClientError(msg)
        return uuid

    @classmethod
    def add_identity(cls, db, identity, backend):
        """ Load and identity list from backend in Sorting Hat """
        uuid = None

        try:
            uuid = cls.add_id(db, backend, email=identity['email'],
                              name=identity['name'], username=identity['username'])
        except UnicodeEncodeError:
            logger.warning("[sortinghat] UnicodeEncodeError. Ignoring it. {} {} {}".format(
                           identity['email'], identity['name'], identity['username']))
        except SortingHatClientError as e:
            logger.warning("[sortinghat] Client error adding identity: {}. Ignoring it. {} {} {}".format(
                           e, identity['email'], identity['name'], identity['username']))
        except Exception:
            logger.warning("[sortinghat] Unknown exception adding identity. Ignoring it. {} {} {}".format(
                           identity['email'], identity['name'], identity['username']))

        if 'company' in identity and identity['company'] is not None:
            org_name = cls.add_organization(db, identity['company'])
            logger.debug("[sortinghat] New organization added {}".format(org_name))

            org_name = cls.add_enrollment(db, uuid, identity['company'])
            logger.debug("[sortinghat] New enrollment added {}".format(org_name))

        return uuid

    @classmethod
    def add_identities(cls, db, identities, backend):
        """ Load identities list from backend in Sorting Hat """

        logger.debug("[sortinghat] Adding identities")

        op = Operation(SortingHatSchema.SortingHatMutation)
        for i, identity in enumerate(identities):
            args = {
                "source": backend,
                "email": identity['email'],
                "name": identity['name'],
                "username": identity['username']
            }
            args_without_empty = {k: v for k, v in args.items() if v}
            if args_without_empty:
                op.add_identity(**args_without_empty, __alias__=f'identity_{i}')
        try:
            db.execute(op)
        except SortingHatClientError as ex:
            for error in ex.errors:
                msg = error['message']
                if 'already exists' in msg:
                    logger.debug("[sortinghat] {}".format(msg))
                else:
                    logger.warning("[sortinghat] {}".format(msg))

    @classmethod
    def add_organization(cls, db, organization):
        try:
            op = Operation(SortingHatSchema.SortingHatMutation)
            add_org = op.add_organization(name=organization)
            add_org.organization().name()
            result = db.execute(op)
            org_name = result['data']['addOrganization']['organization']['name']
        except SortingHatClientError as e:
            logger.error("[sortinghat] Error add organization {}: {}".format(organization, e.errors[0]['message']))
            org_name = None

        return org_name

    @classmethod
    def do_affiliate(cls, db):
        entities = SortingHat.unique_identities(db)
        if not entities:
            return
        mks = [e['mk'] for e in entities]
        args = {
            "uuids": mks
        }
        try:
            op = Operation(SortingHatSchema.SortingHatMutation)
            op.affiliate(**args).job_id
            result = db.execute(op)
            job_id = result['data']['affiliate']['jobId']
            logger.info("[sortinghat] Affiliate job id: {}".format(job_id))
            wait = cls.check_job(db, job_id)
            while wait:
                time.sleep(SLEEP_TIME)
                wait = cls.check_job(db, job_id)
            logger.info("[sortinghat] Affiliate finished job id: {}".format(job_id))
        except SortingHatClientError as e:
            logger.error("[sortinghat] Error affiliate job id: {}\n{}".format(job_id, e.errors[0]))

        return

    @classmethod
    def do_autogender(cls):
        return None

    @classmethod
    def do_unify(cls, db, kwargs):
        entities = SortingHat.unique_identities(db)
        if not entities:
            return
        mks = [e['mk'] for e in entities]
        try:
            op = Operation(SortingHatSchema.SortingHatMutation)
            args = {
                'criteria': [kwargs['matching']],
                'source_uuids': mks
            }
            op.unify(**args).job_id
            result = db.execute(op)
            job_id = result['data']['unify']['jobId']
            logger.info("[sortinghat] Unify job id: {}".format(job_id))
            wait = cls.check_job(db, job_id)
            while wait:
                time.sleep(SLEEP_TIME)
                wait = cls.check_job(db, job_id)
        except SortingHatClientError as e:
            logger.error("[sortinghat] Error unify criteria: {}\n{}".format(args['criteria'], e.errors[0]))

        return

    @classmethod
    def remove_identity(cls, sh_db, ident_id):
        """Delete an identity or unique identity from SortingHat.

        :param sh_db: SortingHat database
        :param ident_id: identity identifier
        """
        success = False
        try:
            op = Operation(SortingHatSchema.SortingHatMutation)
            delete = op.delete_identity(uuid=ident_id)
            delete.uuid()
            result = sh_db.execute(op)
            removed_uuid = result['data']['deleteIdentity']['uuid']
            logger.debug("[sortinghat] Identity {} deleted".format(removed_uuid))
            success = True
        except Exception as e:
            logger.error("[sortinghat] Error remove identity {}: {}".format(ident_id, e))

        return success

    @classmethod
    def update_profile(cls, db, uuid, data):
        args = {
            'uuid': uuid,
            'data': data
        }
        try:
            op = Operation(SortingHatSchema.SortingHatMutation)
            update = op.update_profile(**args)
            update.uuid()
            result = db.execute(op)
            uuid = result['data']['updateProfile']['uuid']
            logger.debug("[sortinghat] updated: {}".format(uuid))
        except SortingHatClientError as ex:
            msg = ex.errors[0]['message']
            logger.error("[sortinghat] updating {}: {}".format(args, msg))

    # Queries
    @classmethod
    def check_job(cls, db, id):
        args = {
            "job_id": id
        }
        wait = True

        op = Operation(SortingHatSchema.Query)
        jobid = op.job(**args)
        jobid.status()
        jobid.errors()
        result = db.execute(op)
        job = result['data']['job']
        if job['status'] == 'finished':
            wait = False
        elif job['status'] == 'failed':
            error_msg = {
                "msg": "job failed",
                "errors": job['errors']
            }
            raise SortingHatClientError(**error_msg)

        return wait

    @classmethod
    def is_bot(cls, db, uuid):
        bot = None
        args = {
            'filters': {
                'uuid': uuid
            }
        }
        try:
            op = Operation(SortingHatSchema.Query)
            op.individuals(**args)
            individual = op.individuals().entities()
            profile = individual.profile()
            profile.is_bot()
            result = db.execute(op)
            bot = result['data']['individuals']['entities'][0]['profile']['isBot']
        except SortingHatClientError as e:
            logger.error("[sortinghat] Error is bot {}: {}".format(uuid, e.errors[0]['message']))
        return bot

    @classmethod
    def get_enrollments(cls, db, uuid):
        args = {
            'page': PAGE,
            'page_size': PAGE_SIZE,
            'filters': {
                'uuid': uuid
            }
        }
        try:
            op = Operation(SortingHatSchema.Query)
            op.individuals(**args)
            individual = op.individuals().entities()
            enrollments = individual.enrollments()
            enrollments.group().name()
            enrollments.group().type()
            enrollments.group().parent_org().name()
            enrollments.start()
            enrollments.end()
            result = db.execute(op)
            enroll = result['data']['individuals']['entities'][0]['enrollments']
        except SortingHatClientError as e:
            logger.error("[sortinghat] Error get enrollments {}: {}".format(uuid, e.errors[0]['message']))
            enroll = None

        return enroll

    @classmethod
    def get_entity(cls, db, id):
        entity = None
        args = {
            'filters': {
                'uuid': id
            }
        }
        try:
            op = Operation(SortingHatSchema.Query)
            op.individuals(**args)
            individual = op.individuals().entities()
            individual.mk()
            identities = individual.identities()
            identities.uuid()
            identities.name()
            identities.email()
            identities.username()
            profile = individual.profile()
            profile.name()
            profile.email()
            profile.gender()
            profile.gender_acc()
            profile.is_bot()
            enrollments = individual.enrollments()
            enrollments.group().parent_org().name()
            enrollments.group().name()
            enrollments.group().type()
            enrollments.start()
            enrollments.end()
            result = db.execute(op)
            if result['data']['individuals']['entities']:
                entity = result['data']['individuals']['entities'][0]
        except SortingHatClientError as e:
            logger.error("[sortinghat] Error get entities {}: {}".format(id, e.errors[0]['message']))
        return entity

    @classmethod
    def get_unique_identity(cls, db, uuid):
        args = {
            'filters': {
                'uuid': uuid
            }
        }
        try:
            op = Operation(SortingHatSchema.Query)
            op.individuals(**args)
            individual = op.individuals().entities()
            individual.mk()
            profile = individual.profile()
            profile.name()
            profile.email()
            profile.gender()
            profile.gender_acc()
            result = db.execute(op)
            p = result['data']['individuals']['entities'][0]['profile']
        except SortingHatClientError as e:
            logger.error("[sortinghat] Error get unique identity {}: {}".format(uuid, e.errors[0]['message']))
            p = None

        return p

    @classmethod
    def get_uuid_from_id(cls, db, sh_id):
        uuid = None
        args = {
            "filters": {
                "uuid": sh_id
            }
        }
        try:
            op = Operation(SortingHatSchema.Query)
            op.individuals(**args)
            individual = op.individuals().entities()
            individual.mk()
            result = db.execute(op)
            entities = result['data']['individuals']['entities']
            mks = [e['mk'] for e in entities]
            if mks:
                uuid = mks[0]
        except SortingHatClientError as e:
            logger.error("[sortinghat] Error get uuid from id {}: {}".format(sh_id, e.errors[0]['message']))

        return uuid

    @classmethod
    def get_uuids_from_profile_name(cls, db, profile_name):
        """ Get the uuid for a profile name """

        def fetch_mk(entities, name):
            mks = []
            for e in entities:
                mk = e['mk']
                if e['profile']['name'] == name:
                    mks.append(mk)

            return mks

        args = {
            'page': 1,
            'page_size': 10,
            'filters': {
                'term': profile_name
            }
        }
        try:
            op = Operation(SortingHatSchema.Query)
            op.individuals(**args)
            individual = op.individuals().entities()
            individual.mk()
            individual.profile().name()
            page_info = op.individuals().page_info()
            page_info.has_next()
            result = db.execute(op)
            entities = result['data']['individuals']['entities']
            mks = fetch_mk(entities, profile_name)
            has_next = result['data']['individuals']['pageInfo']['hasNext']
            while has_next:
                page = args['page']
                args['page'] = page + 1
                op = Operation(SortingHatSchema.Query)
                op.individuals(**args)
                individual = op.individuals().entities()
                individual.mk()
                individual.profile().name()
                page_info = op.individuals().page_info()
                page_info.has_next()
                result = db.execute(op)
                entities = result['data']['individuals']['entities']
                mks.extend(fetch_mk(entities, profile_name))
                has_next = result['data']['individuals']['pageInfo']['hasNext']
        except SortingHatClientError as e:
            logger.error("[sortinghat] Error get uuid from profile name {}: {}".
                         format(profile_name, e.errors[0]['message']))

        return mks

    @classmethod
    def unique_identities(cls, sh_db):
        """List the unique identities available in SortingHat.

        :param sh_db: SortingHat database
        """
        args = {
            'page': PAGE,
            'page_size': PAGE_SIZE
        }
        try:
            op = Operation(SortingHatSchema.Query)
            op.individuals(**args)
            individual = op.individuals().entities()
            individual.mk()
            individual.identities().source()
            individual.identities().uuid()
            page_info = op.individuals().page_info()
            page_info.has_next()
            result = sh_db.execute(op)
            entities = result['data']['individuals']['entities']
            has_next = result['data']['individuals']['pageInfo']['hasNext']
            while has_next:
                page = args['page']
                args['page'] = page + 1
                op = Operation(SortingHatSchema.Query)
                op.individuals(**args)
                individual = op.individuals().entities()
                individual.mk()
                individual.identities().source()
                individual.identities().uuid()
                page_info = op.individuals().page_info()
                page_info.has_next()
                result = sh_db.execute(op)
                entities += result['data']['individuals']['entities']
                has_next = result['data']['individuals']['pageInfo']['hasNext']
            return entities
        except SortingHatClientError as e:
            logger.debug("[sortinghat] Error list unique identities: {}".format(e))

    @classmethod
    def search_last_modified_identities(cls, db, after):
        args = {
            'page': PAGE,
            'page_size': PAGE_SIZE,
            'filters': {
                'lastUpdated': '>' + after.strftime('%Y-%m-%dT%H:%M:%S.%f+00:00')
            }
        }
        try:
            has_next = True
            while has_next:
                op = Operation(SortingHatSchema.Query)
                op.individuals(**args)
                individual = op.individuals().entities()
                individual.mk()
                identities = individual.identities()
                identities.uuid()
                identities.name()
                identities.email()
                identities.username()
                profile = individual.profile()
                profile.name()
                profile.email()
                profile.gender()
                profile.gender_acc()
                profile.is_bot()
                enrollments = individual.enrollments()
                enrollments.group().parent_org().name()
                enrollments.group().name()
                enrollments.group().type()
                enrollments.start()
                enrollments.end()
                page_info = op.individuals().page_info()
                page_info.has_next()
                result = db.execute(op)
                args['page'] = args['page'] + 1
                has_next = result['data']['individuals']['pageInfo']['hasNext']
                entities = result['data']['individuals']['entities']
                yield entities
        except SortingHatClientError as e:
            logger.error("[sortinghat] Error searching identities after {}"
                         ": {}".format(after, e.errors[0]['message']))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Alvaro del Castillo <acs@bitergia.com>
#     Valerio Cosentino <valcos@bitergia.com>
#
import logging
import unittest
import time

import requests
from base import TestBaseBackend
from grimoire_elk.raw.mattermost import MattermostOcean
from grimoire_elk.enriched.utils import REPO_LABELS
from grimoire_elk.enriched.enrich import logger


HEADER_JSON = {"Content-Type": "application/json"}


class TestMattermost(TestBaseBackend):
    """Test Mattermost backend"""

    connector = "mattermost"
    ocean_index = "test_" + connector
    enrich_index = "test_" + connector + "_enrich"

    def test_has_identites(self):
        """Test value of has_identities method"""

        enrich_backend = self.connectors[self.connector][2]()
        self.assertTrue(enrich_backend.has_identities())

    def test_items_to_raw(self):
        """Test whether JSON items are properly inserted into ES"""

        result = self._test_items_to_raw()
        self.assertEqual(result['items'], 89)
        self.assertEqual(result['raw'], 89)

    def test_raw_to_enrich(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich()
        self.assertEqual(result['raw'], 89)
        self.assertEqual(result['enrich'], 89)

        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            self.assertEqual(eitem['channel_name'], "Eclipse Che")

    def test_enrich_repo_labels(self):
        """Test whether the field REPO_LABELS is present in the enriched items"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            self.assertIn(REPO_LABELS, eitem)

    def test_raw_to_enrich_sorting_hat(self):
        """Test enrich with SortingHat"""

        result = self._test_raw_to_enrich(sortinghat=True)
        self.assertEqual(result['raw'], 89)
        self.assertEqual(result['enrich'], 89)

        enrich_backend = self.connectors[self.connector][2]()

        url = self.es_con + "/" + self.enrich_index + "/_search"
        response = enrich_backend.requests.get(url, verify=False).json()
        for hit in response['hits']['hits']:
            source = hit['_source']
            if 'author_uuid' in source:
                self.assertIn('author_domain', source)
                self.assertIn('author_gender', source)
                self.assertIn('author_gender_acc', source)
                self.assertIn('author_org_name', source)
                self.assertIn('author_bot', source)
                self.assertIn('author_multi_org_names', source)

    def test_raw_to_enrich_projects(self):
        """Test enrich with Projects"""

        result = self._test_raw_to_enrich(projects=True)
        self.assertEqual(result['raw'], 89)
        self.assertEqual(result['enrich'], 89)

        res = requests.get(self.es_con + "/" + self.enrich_index + "/_search", verify=False)
        for eitem in res.json()['hits']['hits']:
            self.assertEqual(eitem['_source']['project'], "grimoire")

    def test_copy_raw_fields(self):
        """Test copied raw fields"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            for attribute in enrich_backend.RAW_FIELDS_COPY:
                if attribute in item:
                    self.assertEqual(item[attribute], eitem[attribute])
                else:
                    self.assertIsNone(eitem[attribute])

    def test_refresh_identities(self):
        """Test refresh identities"""

        result = self._test_refresh_identities()
        # ... ?

    def test_perceval_params_legacy_url(self):
        """Test the extraction of perceval params from an URL"""

        url = "https://chat.openshift.io 8j366ft5affy3p36987pcugaoa"
        expected_params = [
            'https://chat.openshift.io',
            '8j366ft5affy3p36987pcugaoa'
        ]
        self.assertListEqual(MattermostOcean.get_perceval_params_from_url(url), expected_params)

        url = "https://some-server-with-dashes.dev/ 993bte1an3dyjmqdgxsr8jr5kh"
        expected_params = [
            'https://some-server-with-dashes.dev',
            '993bte1an3dyjmqdgxsr8jr5kh'
        ]
        self.assertListEqual(MattermostOcean.get_perceval_params_from_url(url), expected_params)

        url = "hTtps://some-SERVER-with-dashes.dev/ 993bte1an3DYJMQDGXSR8JR5kh"
        expected_params = [
            'https://some-server-with-dashes.dev',
            '993bte1an3dyjmqdgxsr8jr5kh'
        ]
        self.assertListEqual(MattermostOcean.get_perceval_params_from_url(url), expected_params)

        # Should fail: wrong protocol
        url = "sftp://cool-mattermost.host/ 993bte1an3dyjmqdgxsr8jr5kh"
        with self.assertRaises(RuntimeError):
            MattermostOcean.get_perceval_params_from_url(url)

        # Should fail: wrong length of channel ID
        url = "https://cool-mattermost.host/ this0has0the0wrong0number0of0letters"
        with self.assertRaises(RuntimeError):
            MattermostOcean.get_perceval_params_from_url(url)

        # Should fail: bad characters in channel id
        url = "sftp://cool-mattermost.host/ 993bteεan3yjmqdgxsr8jr5kh"
        with self.assertRaises(RuntimeError):
            MattermostOcean.get_perceval_params_from_url(url)

        # Should fail: wrong number of parameters
        url = "https://cool-mattermost.host/ 993bte1an3dyjmqdgxsr8jr5kh extra_stuff"
        with self.assertRaises(RuntimeError):
            MattermostOcean.get_perceval_params_from_url(url)

    def test_perceval_params_standard_url(self):
        """Test the extraction of perceval params from an URL"""

        url = "https://my.mattermost.host/some_team/channels/some_channel"
        expected_params = [
            'https://my.mattermost.host',
            'some_channel',
            'some_team',
        ]
        self.assertListEqual(MattermostOcean.get_perceval_params_from_url(url), expected_params)

        url = "http://my-cooler-mattermost.host/team2/channels/channel2/"
        expected_params = [
            'http://my-cooler-mattermost.host',
            'channel2',
            'team2',
        ]
        self.assertListEqual(MattermostOcean.get_perceval_params_from_url(url), expected_params)

        url = "httP://my-cooler-maTTERMost.host/TRansFoRM/channels/CAPitALs/"
        expected_params = [
            'http://my-cooler-mattermost.host',
            'capitals',
            'transform',
        ]
        self.assertListEqual(MattermostOcean.get_perceval_params_from_url(url), expected_params)

        # Should fail:  Malformed host
        url = "https://malformed-host/some_team/channels/some_channel"
        with self.assertRaises(RuntimeError):
            MattermostOcean.get_perceval_params_from_url(url)

        # Should fail:  Wrong protocol
        url = "gemini://my.mattermost.host/some_team/channels/some_channel"
        with self.assertRaises(RuntimeError):
            MattermostOcean.get_perceval_params_from_url(url)

        # Should fail:  Wrong endpoint
        url = "https://my.mattermost.host/some_team/wrongnode/some_channel"
        with self.assertRaises(RuntimeError):
            MattermostOcean.get_perceval_params_from_url(url)

        # Should fail:  No channel name
        url = "https://my.mattermost.host/some_team/channels/"
        with self.assertRaises(RuntimeError):
            MattermostOcean.get_perceval_params_from_url(url)

        # Should fail:  No team name
        url = "https://my.mattermost.host/channels/some_channel"
        with self.assertRaises(RuntimeError):
            MattermostOcean.get_perceval_params_from_url(url)

        # Should fail:  Empty team name
        url = "https://my.mattermost.host//channels/some_channel"
        with self.assertRaises(RuntimeError):
            MattermostOcean.get_perceval_params_from_url(url)

        # Should fail:  Bad team name
        url = "https://my.mattermost.host/i put spaces/channels/some_channel"
        with self.assertRaises(RuntimeError):
            MattermostOcean.get_perceval_params_from_url(url)

    def test_demography_study(self):
        """ Test that the demography study works correctly """

        alias = "demographics"
        study, ocean_backend, enrich_backend = self._test_study('enrich_demography')

        with self.assertLogs(logger, level='INFO') as cm:
            if study.__name__ == "enrich_demography":
                study(ocean_backend, enrich_backend, alias)
            self.assertEqual(cm.output[0], 'INFO:grimoire_elk.enriched.enrich:[mattermost] Demography '
                                           'starting study %s/test_mattermost_enrich'
                             % self.es_con)
            self.assertEqual(cm.output[-1], 'INFO:grimoire_elk.enriched.enrich:[mattermost] Demography '
                                            'end %s/test_mattermost_enrich'
                             % self.es_con)

        time.sleep(5)  # HACK: Wait until github enrich index has been written
        items = [item for item in enrich_backend.fetch()]
        self.assertEqual(len(items), 89)

        for item in items:
            if 'author_uuid' in item:
                self.assertTrue('demography_min_date' in item.keys())
                self.assertTrue('demography_max_date' in item.keys())

        r = enrich_backend.elastic.requests.get(enrich_backend.elastic.index_url + "/_alias",
                                                headers=HEADER_JSON, verify=False)
        self.assertIn(alias, r.json()[enrich_backend.elastic.index]['aliases'])


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

from .elastic import ElasticOcean
from ..elastic_mapping import Mapping as BaseMapping


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = '''
         {
            "dynamic":true,
                "properties": {
                    "data": {
                        "properties": {
                            "details": {
                                "properties": {
                                    "suggested_topics": {
                                        "dynamic": false,
                                        "properties": {
                                            "slug": {
                                                "type": "text",
                                                "index": true
                                            },
                                            "title": {
                                                "type": "text",
                                                "index": true
                                            }
                                        }
                                    }
                                }
                            },
                            "fancy_title": {
                                "type": "text",
                                "index": true
                            },
                            "slug": {
                                "type": "text",
                                "index": true
                            },
                            "title": {
                                "type": "text",
                                "index": true
                            },
                            "post_stream": {
                                "dynamic": false,
                                "properties": {
                                    "posts": {
                                        "properties": {
                                            "cooked": {
                                                "type": "text",
                                                "index": true
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
        }
        '''

        return {"items": mapping}


class DiscourseOcean(ElasticOcean):
    """Discourse Ocean feeder"""

    mapping = Mapping



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Quan Zhou <quan@bitergia.com>
#


import configparser
import copy
import json
import logging
import unittest

from grimoire_elk.enriched.sortinghat_gelk import SortingHat
from sgqlc.operation import Operation
from sortinghat.cli.client import SortingHatClient, SortingHatSchema


CONFIG_FILE = 'tests.conf'
REMOTE_IDENTITIES_FILE = 'data/remote_identities_sortinghat.json'


def read_file(filename, mode='r'):
    with open(filename, mode) as f:
        content = f.read()
    return content


class TestSortinghatGelk(unittest.TestCase):
    @staticmethod
    def get_organizations(client):
        args = {
            "page": 1,
            "page_size": 10
        }
        op = Operation(SortingHatSchema.Query)
        org = op.organizations(**args)
        org.entities().name()
        result = client.execute(op)
        organizations = result['data']['organizations']['entities']

        return organizations

    @staticmethod
    def delete_identity(client, args):
        op = Operation(SortingHatSchema.SortingHatMutation)
        identity = op.delete_identity(**args)
        identity.uuid()
        client.execute(op)

    @staticmethod
    def delete_organization(client, args):
        op = Operation(SortingHatSchema.SortingHatMutation)
        org = op.delete_organization(**args)
        org.organization.name()
        client.execute(op)

    @staticmethod
    def add_identity(client, args):
        op = Operation(SortingHatSchema.SortingHatMutation)
        identity = op.add_identity(**args)
        identity.uuid()
        client.execute(op)

    @staticmethod
    def add_organization(client, args):
        op = Operation(SortingHatSchema.SortingHatMutation)
        org = op.add_organization(**args)
        org.organization.name()
        client.execute(op)

    @staticmethod
    def add_domain(client, args):
        op = Operation(SortingHatSchema.SortingHatMutation)
        dom = op.add_domain(**args)
        dom.domain.domain()
        client.execute(op)

    def setUp(self):
        config = configparser.ConfigParser()
        config.read(CONFIG_FILE)

        # Sorting hat settings
        db_user = config.get('Database', 'user', fallback='')
        db_password = config.get('Database', 'password', fallback='')
        db_host = config.get('Database', 'host', fallback='127.0.0.1')
        db_port = config.get('Database', 'port', fallback=None)
        db_path = config.get('Database', 'path', fallback=None)
        db_ssl = config.getboolean('Database', 'ssl', fallback=False)
        db_verify_ssl = config.getboolean('Database', 'verify_ssl', fallback=True)
        db_tenant = config.get('Database', 'tenant', fallback=None)

        self.sh_db = SortingHatClient(host=db_host,
                                      port=db_port,
                                      user=db_user,
                                      password=db_password,
                                      path=db_path,
                                      ssl=db_ssl,
                                      verify_ssl=db_verify_ssl,
                                      tenant=db_tenant)
        self.sh_db.connect()
        # Clean database
        # Remove identities
        entities = SortingHat.unique_identities(self.sh_db)
        if entities:
            mks = [e['mk'] for e in entities]
            for i in mks:
                arg = {
                    'uuid': i
                }
                self.delete_identity(self.sh_db, arg)

            # Remove organization
            organizations = self.get_organizations(self.sh_db)
            for org in organizations:
                self.delete_organization(self.sh_db, org)

        # Load test data
        data = json.loads(read_file("data/task-identities-data.json"))

        identities = data['identities']
        for identity in identities:
            self.add_identity(self.sh_db, identity)

        organizations = data['organizations']
        for org in organizations:
            self.add_organization(self.sh_db, {"name": org['organization']})
            self.add_domain(self.sh_db, org)

    def test_add_enrollment(self):
        uuid = "39cfb450296d57f61bf75b07b9da49123943c111"
        org = "org1"
        self.assertEqual(SortingHat.add_enrollment(self.sh_db, uuid, org), org)

        _ = SortingHat.get_enrollments(self.sh_db, uuid)
        args = {
            'filters': {
                'uuid': uuid
            }
        }
        op = Operation(SortingHatSchema.Query)
        op.individuals(**args)
        individual = op.individuals().entities()
        individual.enrollments().group().name()
        result = self.sh_db.execute(op)
        entity = result['data']['individuals']['entities'][0]
        self.assertEqual(entity['enrollments'][0]['group']['name'], org)

    def test_add_id(self):
        new_identity = {
            'source': 'git',
            'email': 'test@test.com',
            'name': 'test_add_id',
            'username': 'test_add_id'
        }

        args = copy.deepcopy(new_identity)
        args['db'] = self.sh_db

        uuid = SortingHat.add_id(**args)

        ind_args = {
            'filters': {
                'uuid': uuid
            }
        }
        op = Operation(SortingHatSchema.Query)
        op.individuals(**ind_args)
        individual = op.individuals().entities().identities()
        individual.name()
        individual.email()
        individual.username()
        individual.source()
        result = self.sh_db.execute(op)
        identity = result['data']['individuals']['entities'][0]['identities'][0]

        self.assertDictEqual(identity, new_identity)

    def test_add_organization(self):
        new_org = "test_org"
        self.assertEqual(SortingHat.add_organization(self.sh_db, new_org), new_org)

        args = {
            'page': 1,
            'page_size': 1,
            'filters': {
                'name': new_org
            }
        }
        op = Operation(SortingHatSchema.Query)
        op.organizations(**args)
        op.organizations().entities().name()
        result = self.sh_db.execute(op)
        org = result['data']['organizations']['entities'][0]['name']
        self.assertEqual(org, new_org)

    def test_update_profile(self):
        uuid = "39cfb450296d57f61bf75b07b9da49123943c111"
        origin_args = {
            'filters': {
                'uuid': uuid
            }
        }
        op = Operation(SortingHatSchema.Query)
        op.individuals(**origin_args)
        individual = op.individuals().entities().profile()
        individual.name()
        individual.email()
        result = self.sh_db.execute(op)
        origin_identity = result['data']['individuals']['entities'][0]['profile']

        update = {
            'name': 'new_name',
            'email': 'new@new.com'
        }
        SortingHat.update_profile(self.sh_db, uuid, update)
        op = Operation(SortingHatSchema.Query)
        op.individuals(**origin_args)
        individual = op.individuals().entities().profile()
        individual.name()
        individual.email()
        result = self.sh_db.execute(op)
        new_identity = result['data']['individuals']['entities'][0]['profile']

        self.assertDictEqual(new_identity, update)
        self.assertNotEqual(new_identity['name'], origin_identity['name'])
        self.assertNotEqual(new_identity['email'], origin_identity['email'])

    def test_get_enrollements(self):
        uuid = "39cfb450296d57f61bf75b07b9da49123943c111"
        org = "org1"
        self.assertEqual(SortingHat.add_enrollment(self.sh_db, uuid, org), org)
        enroll = SortingHat.get_enrollments(self.sh_db, uuid)

        self.assertEqual(enroll[0]['group']['name'], org)

    def test_get_entity(self):
        from_uuid = "3942d2eb291ec68dee659692f5d07cb2df9c74ce"
        to_uuid = "39cfb450296d57f61bf75b07b9da49123943c111"

        op = Operation(SortingHatSchema.SortingHatMutation)
        merge = op.merge(from_uuids=[from_uuid], to_uuid=to_uuid)
        merge.uuid()
        self.sh_db.execute(op)

        entity = SortingHat.get_entity(self.sh_db, from_uuid)
        self.assertEqual(entity['mk'], to_uuid)
        self.assertEqual(len(entity['identities']), 2)

    def test_get_unique_identity(self):
        uuid = "39cfb450296d57f61bf75b07b9da49123943c111"
        name = 'user9'
        email = 'user9@org2.com'
        unique_identity = SortingHat.get_unique_identity(self.sh_db, uuid)
        self.assertEqual(unique_identity['name'], name)
        self.assertEqual(unique_identity['email'], email)

    def test_get_uuid_from_id(self):
        from_uuid = "3942d2eb291ec68dee659692f5d07cb2df9c74ce"
        to_uuid = "39cfb450296d57f61bf75b07b9da49123943c111"

        op = Operation(SortingHatSchema.SortingHatMutation)
        merge = op.merge(from_uuids=[from_uuid], to_uuid=to_uuid)
        merge.uuid()
        self.sh_db.execute(op)

        uuid = SortingHat.get_uuid_from_id(self.sh_db, from_uuid)
        self.assertEqual(uuid, to_uuid)

    def test_is_bot(self):
        uuid = "39cfb450296d57f61bf75b07b9da49123943c111"
        self.assertFalse(SortingHat.is_bot(self.sh_db, uuid))

        # Update the profile and set "is_bot = True"
        update = {
            'is_bot': True,
        }
        SortingHat.update_profile(self.sh_db, uuid, update)
        self.assertTrue(SortingHat.is_bot(self.sh_db, uuid))

    def test_remove_identity(self):
        uuid = "39cfb450296d57f61bf75b07b9da49123943c111"
        entities_before = SortingHat.unique_identities(self.sh_db)

        SortingHat.remove_identity(self.sh_db, uuid)
        entities_after = SortingHat.unique_identities(self.sh_db)

        mks_before = [mk['mk'] for mk in entities_before]
        mks_after = [mk['mk'] for mk in entities_after]

        self.assertLess(len(entities_after), len(entities_before))
        self.assertIn(uuid, mks_before)
        self.assertNotIn(uuid, mks_after)

    def test_unique_identities(self):
        entities_before = SortingHat.unique_identities(self.sh_db)

        from_uuid = "3942d2eb291ec68dee659692f5d07cb2df9c74ce"
        to_uuid = "39cfb450296d57f61bf75b07b9da49123943c111"
        op = Operation(SortingHatSchema.SortingHatMutation)
        merge = op.merge(from_uuids=[from_uuid], to_uuid=to_uuid)
        merge.uuid()
        self.sh_db.execute(op)

        entities_after = SortingHat.unique_identities(self.sh_db)

        mks_before = [mk['mk'] for mk in entities_before]
        mks_after = [mk['mk'] for mk in entities_after]

        self.assertLess(len(entities_after), len(entities_before))
        self.assertIn(from_uuid, mks_before)
        self.assertNotIn(from_uuid, mks_after)

    def test_get_uuids_from_profile_name(self):
        user = "user9"
        mks = SortingHat.get_uuids_from_profile_name(self.sh_db, user)
        expected_uuid = "39cfb450296d57f61bf75b07b9da49123943c111"
        self.assertEqual(mks[0], expected_uuid)

        updated_uuid = "4fa3da3bd918171fd667592b7bdebc6438768b60"
        update = {
            'name': 'user9'
        }
        SortingHat.update_profile(self.sh_db, updated_uuid, update)

        mks = SortingHat.get_uuids_from_profile_name(self.sh_db, user)
        self.assertEqual(len(mks), 2)
        self.assertListEqual(mks, [expected_uuid, updated_uuid])

    def test_do_affiliation(self):
        self.assertIsNone(SortingHat.do_affiliate(self.sh_db))

        args = {
            'page': 1,
            'page_size': 10
        }
        op = Operation(SortingHatSchema.Query)
        op.individuals(**args)
        individual = op.individuals().entities()
        individual.profile().name()
        individual.enrollments().group().name()
        result = self.sh_db.execute(op)
        entities = result['data']['individuals']['entities']

        enrolls = {}
        for e in entities:
            name = e['profile']['name']
            enroll = e['enrollments'][0]['group']['name']
            if name in enrolls:
                enrolls[name].append(enroll)
            else:
                enrolls[name] = [enroll]

        expected_enrolls = {
            "user1": ["org1"],
            "user2": ["org2"],
            "user3": ["org1"],
            "user4": ["org2", "org3"],
            "user6": ["org6"],
            "user7": ["org1"],
            "user8": ["org8"],
            "user9": ["org2"],
            "user10": ["org3"]
        }

        self.assertDictEqual(enrolls, expected_enrolls)

    def test_do_unify(self):
        kwargs = {
            'matching': "email",
            'fast_matching': True
        }
        self.assertIsNone(SortingHat.do_unify(self.sh_db, kwargs))

        args = {
            'page': 1,
            'page_size': 10
        }
        op = Operation(SortingHatSchema.Query)
        op.individuals(**args)
        individual = op.individuals().entities()
        individual.mk()
        result = self.sh_db.execute(op)
        entities = result['data']['individuals']['entities']

        self.assertEqual(len(entities), 9)

    def test_do_unify_wrong_criteria(self):
        kwargs = {
            'matching': "email-name",
            'fast_matching': True
        }
        with self.assertLogs() as captured:
            SortingHat.do_unify(self.sh_db, kwargs)
        expected_log = "[sortinghat] Error unify criteria: ['email-name']"
        first_line_error = captured.records[1].getMessage().split("\n")[0]
        self.assertEqual(first_line_error, expected_log)


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Alvaro del Castillo <acs@bitergia.com>
#     Valerio Cosentino <valcos@bitergia.com>
#     Quan Zhou <quan@bitergia.com>
#


import datetime
import logging
import os
import requests
import time
import shutil
import unittest

from sgqlc.operation import Operation

from base import TestBaseBackend
from grimoire_elk.raw.git import GitOcean
from grimoire_elk.enriched.enrich import (logger,
                                          anonymize_url)
from grimoire_elk.enriched.git import logger as git_logger
from grimoire_elk.enriched.utils import REPO_LABELS
from sortinghat.cli.client import SortingHatSchema, SortingHatClientError


CONF_FILE = 'tests_sh.cfg'
HEADER_JSON = {"Content-Type": "application/json"}


class TestGit(TestBaseBackend):
    """Test Git backend"""

    connector = "git"
    ocean_index = "test_" + connector
    ocean_index_anonymized = "test_" + connector + "_anonymized"
    ocean_aliases = ["a", "b", "git-raw"]
    enrich_index = "test_" + connector + "_enrich"
    enrich_index_anonymized = "test_" + connector + "_enrich_anonymized"
    enrich_aliases = ["c", "d", "git"]

    def test_has_identites(self):
        """Test value of has_identities method"""

        enrich_backend = self.connectors[self.connector][2]()
        self.assertTrue(enrich_backend.has_identities())

    def test_items_to_raw(self):
        """Test whether JSON items are properly inserted into ES"""

        result = self._test_items_to_raw()
        self.assertEqual(result['items'], 11)
        self.assertEqual(result['raw'], 11)

        aliases = self.ocean_backend.elastic.list_aliases()
        self.assertListEqual(self.ocean_aliases, list(aliases.keys()))

        url = self.es_con + "/" + self.ocean_index + "/_search?size=20"
        response = self.ocean_backend.requests.get(url, verify=False).json()

        time.sleep(5)  # HACK: Wait until git enrich index has been written
        for hit in response['hits']['hits']:
            item = hit['_source']
            self.assertNotIn('username:password', item['origin'])
            self.assertNotIn('username:password', item['tag'])

    def test_raw_to_enrich(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich()
        self.assertEqual(result['raw'], 11)
        self.assertEqual(result['enrich'], 11)

        enrich_backend = self.connectors[self.connector][2]()

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['committer_name'], '')
        self.assertNotIn('username:password', eitem['origin'])
        self.assertNotIn('username:password', eitem['tag'])

        for item in self.items[1:]:
            eitem = enrich_backend.get_rich_item(item)
            self.assertNotEqual(eitem['committer_name'], 'Unknown')
            self.assertNotEqual(eitem['author_name'], 'Unknown')
            self.assertNotIn('username:password', eitem['origin'])
            self.assertNotIn('username:password', eitem['tag'])

        item = self.items[1]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['author_date'], '2012-08-14T14:32:15')
        self.assertEqual(eitem['author_date_weekday'], 2)
        self.assertEqual(eitem['author_date_hour'], 14)
        self.assertEqual(eitem['utc_author_date_weekday'], 2)
        self.assertEqual(eitem['utc_author_date_hour'], 17)
        self.assertEqual(eitem['author_uuid'], '8b8d552af706acff79df0f18f5295391c51acd79')
        self.assertEqual(eitem['author_domain'], 'gmail.com')
        self.assertEqual(eitem['author_name'], 'Eduardo Morais and Zhongpeng Lin')
        self.assertEqual(eitem['commit_date'], '2012-08-14T14:32:15')
        self.assertEqual(eitem['commit_date_weekday'], 2)
        self.assertEqual(eitem['commit_date_hour'], 14)
        self.assertEqual(eitem['utc_commit_date_weekday'], 2)
        self.assertEqual(eitem['utc_commit_date_hour'], 17)
        self.assertListEqual(eitem['acked_by_multi_names'], [])
        self.assertListEqual(eitem['co_developed_by_multi_names'], [])
        self.assertListEqual(eitem['reported_by_multi_names'], [])
        self.assertListEqual(eitem['reviewed_by_multi_names'], [])
        self.assertListEqual(eitem['signed_off_by_multi_names'], [])
        self.assertListEqual(eitem['suggested_by_multi_names'], [])
        self.assertListEqual(eitem['tested_by_multi_names'], [])
        self.assertListEqual(eitem['non_authored_signed_off_by_multi_names'], [])
        self.assertListEqual(eitem['approved_by_multi_names'], [])
        self.assertListEqual(eitem['approved_by_multi_domains'], [])
        self.assertListEqual(eitem['approved_by_multi_bots'], [])
        self.assertListEqual(eitem['approved_by_multi_org_names'], [])
        self.assertListEqual(eitem['approved_by_multi_uuids'], [])
        self.assertListEqual(eitem['co_authored_by_multi_names'], [])
        self.assertListEqual(eitem['co_authored_by_multi_domains'], [])
        self.assertListEqual(eitem['merged_by_multi_names'], [])
        self.assertListEqual(eitem['merged_by_multi_domains'], [])
        self.assertListEqual(eitem['non_authored_approved_by_multi_names'], [])
        self.assertListEqual(eitem['non_authored_approved_by_multi_domains'], [])
        self.assertListEqual(eitem['non_authored_co_authored_by_multi_names'], [])
        self.assertListEqual(eitem['non_authored_co_authored_by_multi_domains'], [])
        self.assertListEqual(eitem['non_authored_merged_by_multi_names'], [])
        self.assertListEqual(eitem['non_authored_merged_by_multi_domains'], [])

        item = self.items[8]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['author_date'], '2014-02-11T22:10:39')
        self.assertEqual(eitem['author_date_weekday'], 2)
        self.assertEqual(eitem['author_date_hour'], 22)
        self.assertEqual(eitem['utc_author_date_weekday'], 3)
        self.assertEqual(eitem['utc_author_date_hour'], 6)
        self.assertEqual(eitem['author_uuid'], 'c77da8cd8adf8cb0d41ee16e57626528c0bf6740')
        self.assertEqual(eitem['author_domain'], 'boss.io')
        self.assertEqual(eitem['author_name'], 'Owl Capone')
        self.assertEqual(eitem['commit_date'], '2014-02-11T22:10:39')
        self.assertEqual(eitem['commit_date_weekday'], 2)
        self.assertEqual(eitem['commit_date_hour'], 22)
        self.assertEqual(eitem['utc_commit_date_weekday'], 3)
        self.assertEqual(eitem['utc_commit_date_hour'], 6)
        self.assertListEqual(eitem['acked_by_multi_names'], [])
        self.assertListEqual(eitem['co_developed_by_multi_names'], [])
        self.assertListEqual(eitem['reported_by_multi_names'], [])
        self.assertListEqual(eitem['reviewed_by_multi_domains'], ['boss.io'])
        self.assertListEqual(eitem['reviewed_by_multi_names'], ['Owl Capone'])
        self.assertListEqual(eitem['signed_off_by_multi_domains'], ['boss.io', 'second.io'])
        self.assertListEqual(eitem['signed_off_by_multi_names'], ['Owl Capone', 'Owl Second'])
        self.assertListEqual(eitem['suggested_by_multi_names'], [])
        self.assertListEqual(eitem['tested_by_multi_names'], [])
        self.assertListEqual(eitem['non_authored_signed_off_by_multi_names'], ['Owl Second'])
        self.assertListEqual(eitem['approved_by_multi_names'], ['Owl Third'])
        self.assertListEqual(eitem['approved_by_multi_domains'], ['third.io'])
        self.assertListEqual(eitem['approved_by_multi_bots'], [False])
        self.assertListEqual(eitem['approved_by_multi_org_names'], ['Unknown'])
        self.assertListEqual(eitem['approved_by_multi_uuids'], ['1bdac6c64760c1027a3777c9ba22adddb0e97c5e'])
        self.assertListEqual(eitem['co_authored_by_multi_names'], ['OwlBot'])
        self.assertListEqual(eitem['co_authored_by_multi_domains'], ['boss.io'])
        self.assertListEqual(eitem['merged_by_multi_names'], ['Owl Third'])
        self.assertListEqual(eitem['merged_by_multi_domains'], ['third.io'])
        self.assertListEqual(eitem['non_authored_approved_by_multi_names'], ['Owl Third'])
        self.assertListEqual(eitem['non_authored_approved_by_multi_domains'], ['third.io'])
        self.assertListEqual(eitem['non_authored_co_authored_by_multi_names'], ['OwlBot'])
        self.assertListEqual(eitem['non_authored_co_authored_by_multi_domains'], ['boss.io'])
        self.assertListEqual(eitem['non_authored_merged_by_multi_names'], ['Owl Third'])
        self.assertListEqual(eitem['non_authored_merged_by_multi_domains'], ['third.io'])

        aliases = self.enrich_backend.elastic.list_aliases()
        self.assertListEqual(self.enrich_aliases, list(aliases.keys()))

    def test_raw_to_enrich_pair_programming(self):
        """Test whether the raw index is properly enriched with pair programming info"""

        result = self._test_raw_to_enrich(pair_programming=True)
        self.assertEqual(result['raw'], 11)
        self.assertEqual(result['enrich'], 16)

        enrich_backend = self.connectors[self.connector][2](pair_programming=True)
        url = self.es_con + "/" + self.enrich_index + "/_search?size=20"
        response = enrich_backend.requests.get(url, verify=False).json()

        time.sleep(5)  # HACK: Wait until git enrich index has been written
        for hit in response['hits']['hits']:
            source = hit['_source']
            self.assertIn('author_date', source)
            self.assertIn('author_date_weekday', source)
            self.assertIn('author_date_hour', source)
            self.assertIn('utc_author_date_weekday', source)
            self.assertIn('utc_author_date_hour', source)
            self.assertIn('author_uuid', source)
            self.assertIn('author_domain', source)
            self.assertIn('author_name', source)
            self.assertIn('commit_date', source)
            self.assertIn('commit_date_weekday', source)
            self.assertIn('commit_date_hour', source)
            self.assertIn('utc_commit_date_weekday', source)
            self.assertIn('utc_commit_date_hour', source)
            self.assertIn('pair_programming_commit', source)
            self.assertIn('pair_programming_files', source)
            self.assertIn('pair_programming_lines_added', source)
            self.assertIn('pair_programming_lines_removed', source)
            self.assertIn('pair_programming_lines_changed', source)
            self.assertIn('is_git_commit_multi_author', source)
            self.assertIn('Signed-off-by_number', source)
            self.assertIn('is_git_commit_signed_off', source)
            self.assertIn('git_uuid', source)

    def test_enrich_repo_labels(self):
        """Test whether the field REPO_LABELS is present in the enriched items"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            self.assertIn(REPO_LABELS, eitem)
            self.assertNotIn('username:password', eitem['origin'])
            self.assertNotIn('username:password', eitem['tag'])

    def test_raw_to_enrich_sorting_hat(self):
        """Test enrich with SortingHat"""

        result = self._test_raw_to_enrich(sortinghat=True)
        self.assertEqual(result['raw'], 11)
        self.assertEqual(result['enrich'], 11)

        enrich_backend = self.connectors[self.connector][2]()
        enrich_backend.sortinghat = True

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertIn('Commit_org_name', eitem)
        self.assertIn('Commit_user_name', eitem)
        self.assertIn('Commit_name', eitem)
        self.assertIn('committer_name', eitem)
        self.assertIn('Author_name', eitem)
        self.assertIn('author_name', eitem)
        self.assertIn('Commit_multi_org_names', eitem)
        self.assertIn('Author_user_name', eitem)
        self.assertIn('Author_multi_org_names', eitem)
        self.assertNotIn('username:password', eitem['origin'])
        self.assertNotIn('username:password', eitem['tag'])

    def test_raw_to_enrich_projects(self):
        """Test enrich with Projects"""

        result = self._test_raw_to_enrich(projects=True)
        self.assertEqual(result['raw'], 11)
        self.assertEqual(result['enrich'], 11)
        enrich_backend = self.connectors[self.connector][2]()
        url = self.es_con + "/" + self.enrich_index + "/_search?size=20"
        response = enrich_backend.requests.get(url, verify=False).json()

        for hit in response['hits']['hits']:
            source = hit['_source']
            if 'author_uuid' in source:
                self.assertIn('author_date', source)
                self.assertIn('author_date_weekday', source)
                self.assertIn('author_date_hour', source)
                self.assertIn('utc_author_date_weekday', source)
                self.assertIn('utc_author_date_hour', source)
                self.assertIn('author_uuid', source)
                self.assertIn('author_domain', source)
                self.assertIn('author_name', source)
                self.assertIn('commit_date', source)
                self.assertIn('commit_date_weekday', source)
                self.assertIn('commit_date_hour', source)
                self.assertIn('utc_commit_date_weekday', source)
                self.assertIn('utc_commit_date_hour', source)
                self.assertIn('project', source)
                self.assertIn('project_1', source)
                self.assertEqual(source['project'], 'Main')
                self.assertEqual(source['project_1'], 'Main')
                self.assertNotIn('username:password', source['origin'])
                self.assertNotIn('username:password', source['tag'])

    def test_raw_to_enrich_sorting_hat_multi_enroll(self):
        """Test enrich with SortingHat with multiple enrollments"""

        def add_individual(task, args):
            op = Operation(SortingHatSchema.SortingHatMutation)
            identity = op.add_identity(**args)
            identity.uuid()
            try:
                result = task.execute(op)
                uuid = result['data']['addIdentity']['uuid']
            except SortingHatClientError as e:
                uuid = e.errors[0]['message'].split("'")[1]
            return uuid

        def add_organization(task, args):
            op = Operation(SortingHatSchema.SortingHatMutation)
            org = op.add_organization(**args)
            org.organization.name()
            try:
                task.execute(op)
            except SortingHatClientError:
                pass

        def add_enroll(task, args):
            op = Operation(SortingHatSchema.SortingHatMutation)
            add_org = op.enroll(**args)
            add_org.individual().enrollments().group().name()
            try:
                task.execute(op)
            except SortingHatClientError:
                pass

        self.enrich_backend = self.connectors[self.connector][2](db_sortinghat="test_sh",
                                                                 db_user=self.db_user,
                                                                 db_password=self.db_password,
                                                                 db_host=self.db_host,
                                                                 db_port=self.db_port,
                                                                 db_path=self.db_path,
                                                                 db_ssl=self.db_ssl,
                                                                 db_verify_ssl=self.db_verify_ssl,
                                                                 db_tenant=self.db_tenant)
        task = self.enrich_backend.sh_db
        individual_args = {
            "name": "Owl Capone",
            "email": "owlcapone@boss.io",
            "source": "git"
        }
        uuid = add_individual(task, individual_args)
        org1_args = {
            "name": "Test Boss"
        }
        org2_args = {
            "name": "Test Second"
        }
        add_organization(task, org1_args)
        add_organization(task, org2_args)
        enroll1_args = {
            "from_date": "1900-01-01T00:00:00",
            "group": "Test Boss",
            "to_date": "2100-01-01T00:00:00",
            "uuid": uuid
        }
        enroll2_args = {
            "from_date": "1900-01-01T00:00:00",
            "group": "Test Second",
            "to_date": "2100-01-01T00:00:00",
            "uuid": uuid
        }
        add_enroll(task, enroll1_args)
        add_enroll(task, enroll2_args)
        result = self._test_raw_to_enrich(sortinghat=True)
        self.assertEqual(result['raw'], 11)
        self.assertEqual(result['enrich'], 11)

        enrich_backend = self.connectors[self.connector][2]()
        enrich_backend.sortinghat = True
        item = self.items[8]
        eitem = enrich_backend.get_rich_item(item)
        self.assertListEqual(eitem['author_multi_org_names'], ['Test Boss', 'Test Second'])
        self.assertListEqual(eitem['signed_off_by_multi_names'], ['Owl Capone', 'Owl Second'])
        self.assertListEqual(eitem['signed_off_by_multi_org_names'], ['Test Boss', 'Test Second', 'Unknown'])

    def test_refresh_identities(self):
        """Test refresh identities"""

        result = self._test_refresh_identities()
        # ... ?

    def test_demography_study(self):
        """ Test that the demography study works correctly """

        alias = 'demographics'
        study, ocean_backend, enrich_backend = self._test_study('enrich_demography')

        with self.assertLogs(logger, level='INFO') as cm:

            if study.__name__ == "enrich_demography":
                study(ocean_backend, enrich_backend, alias, date_field="utc_commit")

            self.assertEqual(cm.output[0], 'INFO:grimoire_elk.enriched.enrich:[git] Demography '
                                           'starting study %s/test_git_enrich'
                             % anonymize_url(self.es_con))
            self.assertEqual(cm.output[-1], 'INFO:grimoire_elk.enriched.enrich:[git] Demography '
                                            'end %s/test_git_enrich'
                             % anonymize_url(self.es_con))

        time.sleep(5)  # HACK: Wait until git enrich index has been written
        items = [item for item in enrich_backend.fetch()]
        self.assertEqual(len(items), 11)
        for item in items:
            self.assertTrue('demography_min_date' in item.keys())
            self.assertTrue('demography_max_date' in item.keys())
            self.assertNotIn('username:password', item['origin'])
            self.assertNotIn('username:password', item['tag'])

        r = enrich_backend.elastic.requests.get(enrich_backend.elastic.index_url + "/_alias",
                                                headers=HEADER_JSON, verify=False)
        self.assertIn(alias, r.json()[enrich_backend.elastic.index]['aliases'])

    def test_extra_study(self):
        """ Test that the extra study works correctly """

        study, ocean_backend, enrich_backend = self._test_study('enrich_extra_data')

        with self.assertLogs(logger, level='INFO') as cm:

            if study.__name__ == "enrich_extra_data":
                study(ocean_backend, enrich_backend,
                      json_url="https://gist.githubusercontent.com/valeriocos/893f55c28c4bd8fa7a217c4e201f4698/raw/"
                               "ba298a6fb09558e68c5e4ec6ae23b1c89fe920ef/test_extra_study.txt")

        time.sleep(5)  # HACK: Wait until git enrich index has been written
        items = [item for item in enrich_backend.fetch()]
        self.assertEqual(len(items), 11)
        for item in items:
            if item['origin'] == '/tmp/perceval_mc84igfc/gittest':
                self.assertIn('extra_secret_repo', item.keys())
            else:
                self.assertNotIn('extra_secret_repo', item.keys())

    def test_enrich_forecast_activity(self):
        """ Test that the forecast activity study works correctly """

        study, ocean_backend, enrich_backend = self._test_study('enrich_forecast_activity')

        with self.assertLogs(logger, level='INFO') as cm:

            if study.__name__ == "enrich_forecast_activity":
                study(ocean_backend, enrich_backend, out_index='test_git_study_forecast_activity', observations=2)

                self.assertEqual(cm.output[0], 'INFO:grimoire_elk.enriched.enrich:'
                                               '[enrich-forecast-activity] Start study')
                self.assertEqual(cm.output[-1], 'INFO:grimoire_elk.enriched.enrich:'
                                                '[enrich-forecast-activity] End study')

        time.sleep(5)  # HACK: Wait until git enrich index has been written
        url = self.es_con + "/test_git_study_forecast_activity/_search?size=20"
        response = enrich_backend.requests.get(url, verify=False).json()
        for hit in response['hits']['hits']:
            source = hit['_source']
            self.assertIn('uuid', source)
            self.assertIn('origin', source)
            self.assertIn('repository', source)
            self.assertIn('interval_months', source)
            self.assertIn('from_date', source)
            self.assertIn('to_date', source)
            self.assertIn('study_creation_date', source)
            self.assertIn('grimoire_creation_date', source)
            self.assertIn('is_git_survived', source)
            self.assertIn('author_uuid', source)
            self.assertIn('author_name', source)
            self.assertIn('author_bot', source)
            self.assertIn('author_user_name', source)
            self.assertIn('author_org_name', source)
            self.assertIn('author_domain', source)
            self.assertIn('prediction_05', source)
            self.assertIn('prediction_07', source)
            self.assertIn('prediction_09', source)
            self.assertIn('next_activity_05', source)
            self.assertIn('next_activity_07', source)
            self.assertIn('next_activity_09', source)
            self.assertIn('metadata__gelk_version', source)
            self.assertIn('metadata__gelk_backend_name', source)
            self.assertIn('metadata__enriched_on', source)

    def test_onion_study(self):
        """ Test that the onion study works correctly """

        alias = "all_onion"
        study, ocean_backend, enrich_backend = self._test_study('enrich_onion')
        study(ocean_backend, enrich_backend, alias, in_index='test_git_enrich', out_index='test_git_onion')

        url = self.es_con + "/_aliases"
        response = requests.get(url, verify=False).json()
        self.assertTrue('test_git_onion' in response)

        time.sleep(1)

        url = self.es_con + "/test_git_onion/_search?size=50"
        response = requests.get(url, verify=False).json()
        hits = response['hits']['hits']
        self.assertEqual(len(hits), 16)
        for hit in hits:
            source = hit['_source']
            self.assertIn('timeframe', source)
            self.assertIn('author_uuid', source)
            self.assertIn('author_name', source)
            self.assertIn('contributions', source)
            self.assertIn('metadata__timestamp', source)
            self.assertIn('project', source)
            self.assertIn('author_org_name', source)
            self.assertIn('cum_net_sum', source)
            self.assertIn('percent_cum_net_sum', source)
            self.assertIn('onion_role', source)
            self.assertIn('quarter', source)
            self.assertIn('metadata__enriched_on', source)
            self.assertIn('data_source', source)
            self.assertIn('grimoire_creation_date', source)

    def test_enrich_areas_of_code(self):
        """ Test that areas of code works correctly"""

        alias = 'enrich_areas_of_code'
        projects_json_repo = "/tmp/perceval_mc84igfc/gittest"
        projects_json = {
            "project": {
                "git": [
                    "/tmp/perceval_mc84igfc/gittest"
                ]
            }
        }
        prjs_map = {
            "git": {
                "/tmp/perceval_mc84igfc/gittest": "project"
            }
        }

        study, ocean_backend, enrich_backend = self._test_study('enrich_areas_of_code',
                                                                projects_json=projects_json,
                                                                prjs_map=prjs_map,
                                                                projects_json_repo=projects_json_repo)

        study(ocean_backend, enrich_backend, alias, in_index='test_git', out_index='test_git_aoc')
        time.sleep(5)  # HACK: Wait until git area of code has been written
        url = self.es_con + "/test_git_aoc/_search?size=20"
        response = enrich_backend.requests.get(url, verify=False).json()
        hits = response['hits']['hits']
        self.assertEqual(len(hits), 12)
        for hit in hits:
            source = hit['_source']
            self.assertIn('addedlines', source)
            self.assertIn('author_bot', source)
            self.assertIn('author_domain', source)
            self.assertIn('author_id', source)
            self.assertIn('author_name', source)
            self.assertIn('author_org_name', source)
            self.assertIn('author_multi_org_names', source)
            self.assertIn('author_user_name', source)
            self.assertIn('author_uuid', source)
            self.assertIn('committer', source)
            self.assertIn('committer_date', source)
            self.assertIn('date', source)
            self.assertIn('eventtype', source)
            self.assertIn('fileaction', source)
            self.assertIn('filepath', source)
            self.assertIn('files', source)
            self.assertIn('filetype', source)
            self.assertIn('file_name', source)
            self.assertIn('file_ext', source)
            self.assertIn('file_dir_name', source)
            self.assertIn('file_path_list', source)
            self.assertIn('git_author_domain', source)
            self.assertIn('grimoire_creation_date', source)
            self.assertIn('hash', source)
            self.assertIn('id', source)
            self.assertIn('message', source)
            self.assertIn('metadata__enriched_on', source)
            self.assertIn('metadata__timestamp', source)
            self.assertIn('metadata__updated_on', source)
            self.assertIn('origin', source)
            self.assertIn('owner', source)
            self.assertIn('perceval_uuid', source)
            self.assertIn('project', source)
            self.assertIn('project_1', source)
            self.assertIn('removedlines', source)
            self.assertIn('repository', source)
            self.assertIn('uuid', source)
            self.assertEqual(source['origin'], '/tmp/perceval_mc84igfc/gittest')
            self.assertEqual(source['repository'], '/tmp/perceval_mc84igfc/gittest')

    def test_enrich_areas_of_code_extra_fields(self):
        """ Test that areas of code works correctly when the repo contains extra fields"""

        alias = "git_areas_of_code"
        projects_json_repo = "https://github.com/acme/errors"
        projects_json = {
            "labels-repo": {
                "git": [
                    "https://github.com/acme/errors --labels=[errors]"
                ]
            }
        }
        prjs_map = {
            "git": {
                "https://github.com/acme/errors": "labels-repo"
            }
        }

        study, ocean_backend, enrich_backend = self._test_study('enrich_areas_of_code',
                                                                projects_json=projects_json,
                                                                prjs_map=prjs_map,
                                                                projects_json_repo=projects_json_repo)

        study(ocean_backend, enrich_backend, alias, in_index='test_git', out_index='test_git_aoc_repo_name')
        time.sleep(5)  # HACK: Wait until git area of code has been written
        url = self.es_con + "/test_git_aoc_repo_name/_search?size=20"
        response = enrich_backend.requests.get(url, verify=False).json()
        hits = response['hits']['hits']
        self.assertEqual(len(hits), 2)
        for hit in hits:
            source = hit['_source']
            print(source['origin'])
            self.assertEqual(source['origin'], 'https://github.com/acme/errors')
            self.assertEqual(source['repository'], 'https://github.com/acme/errors')

    def test_enrich_areas_of_code_private_repo(self):
        """ Test that areas of code works correctly for git private repos"""

        alias = "git_areas_of_code"
        projects_json_repo = "https://username:password@github.com/acme/errors"
        projects_json = {
            "secret-repo": {
                "git": [
                    "https://username:password@github.com/acme/errors"
                ]
            }
        }
        prjs_map = {
            "git": {
                "https://username:password@github.com/acme/errors": "secret-repo"
            }
        }

        study, ocean_backend, enrich_backend = self._test_study('enrich_areas_of_code',
                                                                projects_json=projects_json,
                                                                prjs_map=prjs_map,
                                                                projects_json_repo=projects_json_repo)

        study(ocean_backend, enrich_backend, alias, in_index='test_git', out_index='test_git_aoc_anonymized')
        time.sleep(5)  # HACK: Wait until git area of code has been written
        url = self.es_con + "/test_git_aoc_anonymized/_search?size=20"
        response = enrich_backend.requests.get(url, verify=False).json()
        hits = response['hits']['hits']
        self.assertEqual(len(hits), 2)
        for hit in hits:
            source = hit['_source']
            self.assertIn('addedlines', source)
            self.assertIn('author_bot', source)
            self.assertIn('author_domain', source)
            self.assertIn('author_id', source)
            self.assertIn('author_name', source)
            self.assertIn('author_org_name', source)
            self.assertIn('author_multi_org_names', source)
            self.assertIn('author_user_name', source)
            self.assertIn('author_uuid', source)
            self.assertIn('committer', source)
            self.assertIn('committer_date', source)
            self.assertIn('date', source)
            self.assertIn('eventtype', source)
            self.assertIn('fileaction', source)
            self.assertIn('filepath', source)
            self.assertIn('files', source)
            self.assertIn('filetype', source)
            self.assertIn('file_name', source)
            self.assertIn('file_ext', source)
            self.assertIn('file_dir_name', source)
            self.assertIn('file_path_list', source)
            self.assertIn('git_author_domain', source)
            self.assertIn('grimoire_creation_date', source)
            self.assertIn('hash', source)
            self.assertIn('id', source)
            self.assertIn('message', source)
            self.assertIn('metadata__enriched_on', source)
            self.assertIn('metadata__timestamp', source)
            self.assertIn('metadata__updated_on', source)
            self.assertIn('origin', source)
            self.assertIn('owner', source)
            self.assertIn('perceval_uuid', source)
            self.assertIn('project', source)
            self.assertIn('project_1', source)
            self.assertIn('removedlines', source)
            self.assertIn('repository', source)
            self.assertIn('uuid', source)
            self.assertEqual(source['origin'], 'https://github.com/acme/errors')
            self.assertEqual(source['repository'], 'https://github.com/acme/errors')

    def test_enrich_git_branches_study(self):
        """ Test that the git branches study works correctly """

        study, ocean_backend, enrich_backend = self._test_study('enrich_git_branches')

        items = [item for item in enrich_backend.fetch()]
        self.assertEqual(len(items), 11)
        for item in items:
            self.assertTrue('branches' in item.keys())

    def test_enrich_git_branches_study_filter_no_collection(self):
        """ Test that the git branches study skip when --filter-no-collection is present """

        projects_json_repo = "https://github.com/grimoirelab/perceval.git --filter-no-collection=true"
        projects_json = {
            "filter-repo": {
                "git": [
                    "https://github.com/grimoirelab/perceval.git --filter-no-collection=true"
                ]
            }
        }
        prjs_map = {
            "git": {
                "https://github.com/grimoirelab/perceval.git --filter-no-collection=true": "filter-repo"
            }
        }

        study, ocean_backend, enrich_backend = self._test_study('enrich_git_branches',
                                                                projects_json=projects_json,
                                                                prjs_map=prjs_map,
                                                                projects_json_repo=projects_json_repo)

        today = datetime.datetime.today().day
        with self.assertLogs(git_logger, level='INFO') as cm:
            if study.__name__ == "enrich_git_branches":
                study(ocean_backend, enrich_backend, run_month_days=[today])
                self.assertEqual(cm.output[0], 'INFO:grimoire_elk.enriched.git:[git] study git-branches start')
                self.assertEqual(cm.output[1], 'INFO:grimoire_elk.enriched.git:[git] study git-branches skipping'
                                               ' repo {}'.format(projects_json_repo))
                self.assertEqual(cm.output[-1], 'INFO:grimoire_elk.enriched.git:[git] study git-branches end')

    def test_enrich_git_branches_study_skip_uncloned_repository(self):
        """ Test that the git branches study skip when the repository is not cloned """

        projects_json_repo = "https://github.com/chaoss/grimoirelab-perceval.git"
        projects_json = {
            "no-cloned": {
                "git": [
                    "https://github.com/chaoss/grimoirelab-perceval.git"
                ]
            }
        }
        prjs_map = {
            "git": {
                "https://github.com/chaoss/grimoirelab-perceval.git": "no-cloned"
            }
        }

        study, ocean_backend, enrich_backend = self._test_study('enrich_git_branches',
                                                                projects_json=projects_json,
                                                                prjs_map=prjs_map,
                                                                projects_json_repo=projects_json_repo)
        # Remove the clone
        base_path = os.path.expanduser('~/.perceval/repositories/')
        processed_uri = projects_json_repo.lstrip('/')
        git_path = os.path.join(base_path, processed_uri) + '-git'
        if os.path.exists(git_path) and os.path.isdir(git_path):
            shutil.rmtree(git_path, ignore_errors=True)

        today = datetime.datetime.today().day
        with self.assertLogs(git_logger, level='INFO') as cm:
            if study.__name__ == "enrich_git_branches":
                study(ocean_backend, enrich_backend, run_month_days=[today])
                self.assertEqual(cm.output[0], 'INFO:grimoire_elk.enriched.git:[git] study git-branches start')
                self.assertEqual(cm.output[1], 'ERROR:grimoire_elk.enriched.git:[git] study git-branches skipping'
                                               ' not cloned repo {}'.format(projects_json_repo))
                self.assertEqual(cm.output[-1], 'INFO:grimoire_elk.enriched.git:[git] study git-branches end')

    def test_perceval_params(self):
        """Test the extraction of perceval params from an URL"""

        url = "https://github.com/grimoirelab/perceval"
        expected_params = [
            'https://github.com/grimoirelab/perceval'
        ]
        self.assertListEqual(GitOcean.get_perceval_params_from_url(url), expected_params)

        url = "https://github.com/grimoirelab/perceval /tmp/perceval-repo"
        expected_params = [
            'https://github.com/grimoirelab/perceval'
        ]
        self.assertListEqual(GitOcean.get_perceval_params_from_url(url), expected_params)

    def test_items_to_raw_anonymized(self):
        """Test whether JSON items are properly inserted into ES anonymized"""

        result = self._test_items_to_raw_anonymized()

        self.assertGreater(result['items'], 0)
        self.assertGreater(result['raw'], 0)
        self.assertEqual(result['items'], result['raw'])

        item = self.items[0]['data']
        self.assertEqual(item['Author'], 'e2ea52f7f782fe08109b762e474ff20656a51f47 <xxxxxx@gmail.com>')
        self.assertEqual(item['Commit'], '')

        item = self.items[1]['data']
        self.assertEqual(item['Author'], '7bb272958a7c0c54de85dc078aa1b98da7b930de <xxxxxx@gmail.com>')
        self.assertEqual(item['Commit'], '7bb272958a7c0c54de85dc078aa1b98da7b930de <xxxxxx@gmail.com>')

        item = self.items[6]['data']
        self.assertEqual(item['Author'], 'abe1a5515d468ed258124c4c946ceb34ef7ffbda <xxxxxx@gmail.com>')
        self.assertEqual(item['Commit'], 'abe1a5515d468ed258124c4c946ceb34ef7ffbda <xxxxxx@gmail.com>')

    def test_raw_to_enrich_anonymized(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich_anonymized()

        self.assertGreater(result['raw'], 0)
        self.assertGreater(result['enrich'], 0)
        self.assertEqual(result['raw'], result['enrich'])

        enrich_backend = self.connectors[self.connector][2]()

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['author_uuid'], '5fe609e6ddadd19697e69832ac1889bb942cccfb')
        self.assertEqual(eitem['author_domain'], 'gmail.com')
        self.assertEqual(eitem['author_name'], 'e2ea52f7f782fe08109b762e474ff20656a51f47')
        self.assertIsNone(eitem['committer_domain'])
        self.assertEqual(eitem['committer_name'], '')

        item = self.items[1]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['author_uuid'], '0442af9edda128740e57331769ee9e68a99ec36f')
        self.assertEqual(eitem['author_domain'], 'gmail.com')
        self.assertEqual(eitem['author_name'], '7bb272958a7c0c54de85dc078aa1b98da7b930de')
        self.assertEqual(eitem['committer_domain'], 'gmail.com')
        self.assertEqual(eitem['committer_name'], '7bb272958a7c0c54de85dc078aa1b98da7b930de')

        item = self.items[6]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['author_uuid'], '221aa5305c73e6574dbba1865ccc930eaa6f59a8')
        self.assertEqual(eitem['author_domain'], 'gmail.com')
        self.assertEqual(eitem['author_name'], 'abe1a5515d468ed258124c4c946ceb34ef7ffbda')
        self.assertEqual(eitem['committer_domain'], 'gmail.com')
        self.assertEqual(eitem['committer_name'], 'abe1a5515d468ed258124c4c946ceb34ef7ffbda')

    def test_copy_raw_fields(self):
        """Test copied raw fields"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            for attribute in enrich_backend.RAW_FIELDS_COPY:
                if attribute in item:
                    self.assertEqual(item[attribute], eitem[attribute])
                else:
                    self.assertIsNone(eitem[attribute])


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')
